/*{ radio_cc1120.c *****************************************************************

 Program Name: cc1120
 
 This source is published under the GNU LESSER GENERAL PUBLIC LICENSE (LGPL).
 See file LEGAL.txt in this software package for details.

 Authors: <Sascha Poggemann>, <Gregor Rebel>

 Description: This is an implementation for the Texas Instruments cc1120 Tranceiver
 
}*/
//{ Private Defines/ TypeDefs ****************************************************

//}  Private Defines
//{ Includes *********************************************************************

#include "radio_cc1120.h"

//} Includes
//{ Private Structures/ Enums ****************************************************

//} Private Structures/ Enums
//{ Global Variables *************************************************************

// required to be called from here
extern void ttc_radio_set_operating_mode(u8_t LogicalIndex, ttc_radio_mode_e Mode);

ttc_array_define(ttc_mutex_smart_t*, TransactionLocks, TTC_AMOUNT_SPIS);

//} Global Variables
//{ Private Function prototypes **************************************************

#define   LockTransaction(SPI_Index, Owner)  while (ttc_mutex_lock_smart( A(TransactionLocks, SPI_Index-1), -1, ( void(*)() ) Owner) != tqe_OK)
#define UnlockTransaction(SPI_Index)         ttc_mutex_unlock_smart( A(TransactionLocks, SPI_Index-1) )
u8_t Amount_NonEmptyPackets = 0; //D

//} Private Function prototypes
//{ Implemented Functions ********************************************************

void radio_cc1120_get_defaults(u8_t RadioIndex, ttc_radio_generic_t* Radio_Generic) {
    Assert_Radio(Radio_Generic, ec_InvalidArgument);
    Radio_Generic->LogicalIndex = RadioIndex;
    radio_cc1120_get_config(Radio_Generic); // will create a driver configuration on first call

    // ToDo: SP - fill Radio_Generic with default values
    Radio_Generic->MaxChannelTx    = cc1120_MAX_Channels;
    Radio_Generic->MaxChannelRx    = cc1120_MAX_Channels;
    Radio_Generic->MaxLevelTx      = cc1120_MAX_TX_LEVEL;
    Radio_Generic->MaxLevelRx      = cc1120_MAX_RX_LEVEL;
    Radio_Generic->Max_Header      = RADIO_cc1120_HEADER_SIZE;  // 2
    Radio_Generic->Max_Footer      = RADIO_cc1120_FOOTER_SIZE;  // 0
    Radio_Generic->Max_Payload     = RADIO_cc1120_PAYLOAD_SIZE; // 60
    Radio_Generic->ChannelRx       = 1;
    Radio_Generic->ChannelTx       = 1;
    Radio_Generic->Size_QueueTx    = 5;
    Radio_Generic->Size_QueueRx    = 5;
    Radio_Generic->TargetAddress   = 0xff; // send to broadcast address
}
void radio_cc1120_get_features(u8_t RadioIndex, ttc_radio_generic_t* Radio_Generic) {
    radio_cc1120_get_config(Radio_Generic); // checks pointer too

    // fill in default values
    radio_cc1120_get_defaults(RadioIndex, Radio_Generic);

    // add max values
    Radio_Generic->Flags.Bit.SleepMode_IDLE = 1;
    Radio_Generic->Flags.Bit.SleepMode_OFF  = 1;
    Radio_Generic->ChannelTx                = cc1120_MAX_Channels;
    Radio_Generic->ChannelRx                = cc1120_MAX_Channels;
    Radio_Generic->LevelTx                  = cc1120_MAX_TX_LEVEL;
    Radio_Generic->LevelRx                  = cc1120_MAX_RX_LEVEL;
}
ttc_radio_errorcode_e radio_cc1120_init(ttc_radio_generic_t* Radio_Generic) {
    if (0) ttc_task_begin_criticalsection(radio_cc1120_init); // disable multitasking during initialization (easier to debug)
    radio_cc1120_driver_t* DriverCfg = radio_cc1120_get_config(Radio_Generic); // checks pointer too
    ttc_radio_errorcode_e Error = tre_OK;
    u8_t State = 0;
    u8_t Level = 0;

    if (! A(TransactionLocks, DriverCfg->SPI_Index-1) )
          A(TransactionLocks, DriverCfg->SPI_Index-1) = ttc_mutex_create_smart();

    LockTransaction(DriverCfg->SPI_Index, radio_cc1120_init);

    // no SPI communication allowed before this point!
    radio_cc1120_spi_init(DriverCfg);

    if (!Error) { Level++; Error = radio_cc1120_spi_get_chip_id(DriverCfg); }
    if (!Error) { Level++; Error = radio_cc1120_spi_get_chip_version(DriverCfg); }
    if (!Error) { Level++; Error = radio_cc1120_spi_reset(DriverCfg); }
    while ( (State = _radio_cc1120_get_state(Radio_Generic)) != rcm_IDLE)
        Radio_Generic->TaskYield();

    // Calibrate radio according to errata
    manualCalibration();
    if (!Error) { Level++; Error = radio_cc1120_init_SmartRF(Radio_Generic); }

    // set RXOFF + TXOFF modes according to cc1120 datasheet p.81
    if (!Error) { Level++; Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_MCSM1, 0x3F); }

    switch (DriverCfg->SPI_Index) { // configure external interrupt line for incoming data
#ifdef TTC_RADIO1
    case 1: {

        // initialize interrupt for receiving data from radio
        ttc_interrupt_init(tit_GPIO_Rising,
                           ttc_gpio_create_index8(TTC_RADIO1_PIN_GDO2),
                           _radio_cc1120_isr_receive,
                           (void*) Radio_Generic,
                           NULL, NULL
                           );
        ttc_interrupt_enable(tit_GPIO_Rising,
                             ttc_gpio_create_index8(TTC_RADIO1_PIN_GDO2),
                             TRUE
                             );
        break;
    }
#endif
#ifdef TTC_RADIO2
    case 2: {

        // initialize interrupt for receiving data from radio
        ttc_interrupt_init(tit_GPIO_Rising,
                           ttc_gpio_create_index8(TTC_RADIO2_PIN_GDO2),
                           _radio_cc1120_isr_receive,
                           (void*) Radio_Generic,
                           NULL, NULL
                           );
        ttc_interrupt_enable(tit_GPIO_Rising,
                             ttc_gpio_create_index8(TTC_RADIO2_PIN_GDO2),
                             TRUE
                             );
        break;
    }
#endif
#ifdef TTC_RADIO3
    case 3: {

        // initialize interrupt for receiving data from radio
        ttc_interrupt_init(tit_GPIO_Rising,
                           ttc_gpio_create_index8(TTC_RADIO3_PIN_GDO2),
                           _radio_cc1120_isr_receive,
                           (void*) Radio_Generic,
                           NULL, NULL
                           );
        ttc_interrupt_enable(tit_GPIO_Rising,
                             ttc_gpio_create_index8(TTC_RADIO3_PIN_GDO2),
                             TRUE
                             );
        break;
    }
#endif
#ifdef TTC_RADIO4
    case 4: {

        // initialize interrupt for receiving data from radio
        ttc_interrupt_init(tit_GPIO_Rising,
                           ttc_gpio_create_index8(TTC_RADIO4_PIN_GDO2),
                           _radio_cc1120_isr_receive,
                           (void*) Radio_Generic,
                           NULL, NULL
                           );
        ttc_interrupt_enable(tit_GPIO_Rising,
                             ttc_gpio_create_index8(TTC_RADIO4_PIN_GDO2),
                             TRUE
                             );
        break;
    }
#endif
#ifdef TTC_RADIO5
    case 5: {

        // initialize interrupt for receiving data from radio
        ttc_interrupt_init(tit_GPIO_Rising,
                           ttc_gpio_create_index8(TTC_RADIO5_PIN_GDO2),
                           _radio_cc1120_isr_receive,
                           (void*) Radio_Generic,
                           NULL, NULL
                           );
        ttc_interrupt_enable(tit_GPIO_Rising,
                             ttc_gpio_create_index8(TTC_RADIO5_PIN_GDO2),
                             TRUE
                             );
        break;
    }
#endif
    default: Assert(0, ec_InvalidConfiguration); // indexed radio not activated
    }

    UnlockTransaction(DriverCfg->SPI_Index);

    // function calls below lock their own transaction
    ttc_radio_set_operating_mode(Radio_Generic->LogicalIndex, trm_Idle);
    if (!Error) { Level++; Error = radio_cc1120_set_power_tx(Radio_Generic, Radio_Generic->LevelTx); }
    ttc_radio_set_operating_mode(Radio_Generic->LogicalIndex, trm_Receive);
    Assert(Error != tre_TimeOut, ec_InvalidConfiguration); // communication error on SPI-bus (check variable Level)
    Assert(Error == tre_OK,      ec_UNKNOWN);              // unknown error (check variable Level)

    if (0) ttc_task_end_criticalsection();
    return tre_OK;
}


// Calibrate radio according to errata
static void manualCalibration(void) {
  
    uint8 original_fs_cal2;
    uint8 calResults_for_vcdac_start_high[3];
    uint8 calResults_for_vcdac_start_mid[3];
    uint8 marcstate;
    uint8 writeByte;
    
    // 1) Set VCO cap-array to 0 (FS_VCO2 = 0x00)
    writeByte = 0x00;
    cc112xSpiWriteReg(CC112X_FS_VCO2, &writeByte, 1);
    
    // 2) Start with high VCDAC (original VCDAC_START + 2):
    cc112xSpiReadReg(CC112X_FS_CAL2, &original_fs_cal2, 1);
    writeByte = original_fs_cal2 + VCDAC_START_OFFSET;
    cc112xSpiWriteReg(CC112X_FS_CAL2, &writeByte, 1);
    
    // 3) Calibrate and wait for calibration to be done (radio back in IDLE state)
    trxSpiCmdStrobe(CC112X_SCAL);
    
    do 
    {
        cc112xSpiReadReg(CC112X_MARCSTATE, &marcstate, 1);
    } while (marcstate != 0x41);
    
    // 4) Read FS_VCO2, FS_VCO4 and FS_CHP register obtained with high VCDAC_START value
    cc112xSpiReadReg(CC112X_FS_VCO2, &calResults_for_vcdac_start_high[FS_VCO2_INDEX], 1);
    cc112xSpiReadReg(CC112X_FS_VCO4, &calResults_for_vcdac_start_high[FS_VCO4_INDEX], 1);
    cc112xSpiReadReg(CC112X_FS_CHP, &calResults_for_vcdac_start_high[FS_CHP_INDEX], 1);
    
    // 5) Set VCO cap-array to 0 (FS_VCO2 = 0x00)
    writeByte = 0x00;
    cc112xSpiWriteReg(CC112X_FS_VCO2, &writeByte, 1);
    
    // 6) Continue with mid VCDAC (original VCDAC_START):
    writeByte = original_fs_cal2;
    cc112xSpiWriteReg(CC112X_FS_CAL2, &writeByte, 1);
    
    // 7) Calibrate and wait for calibration to be done (radio back in IDLE state)
    trxSpiCmdStrobe(CC112X_SCAL);
    
    do 
    {
        cc112xSpiReadReg(CC112X_MARCSTATE, &marcstate, 1);
    } while (marcstate != 0x41);
    
    // 8) Read FS_VCO2, FS_VCO4 and FS_CHP register obtained with mid VCDAC_START value
    cc112xSpiReadReg(CC112X_FS_VCO2, &calResults_for_vcdac_start_mid[FS_VCO2_INDEX], 1);
    cc112xSpiReadReg(CC112X_FS_VCO4, &calResults_for_vcdac_start_mid[FS_VCO4_INDEX], 1);
    cc112xSpiReadReg(CC112X_FS_CHP, &calResults_for_vcdac_start_mid[FS_CHP_INDEX], 1);
    
    // 9) Write back highest FS_VCO2 and corresponding FS_VCO and FS_CHP result
    if (calResults_for_vcdac_start_high[FS_VCO2_INDEX] > calResults_for_vcdac_start_mid[FS_VCO2_INDEX]) 
    {
        writeByte = calResults_for_vcdac_start_high[FS_VCO2_INDEX];
        cc112xSpiWriteReg(CC112X_FS_VCO2, &writeByte, 1);
        writeByte = calResults_for_vcdac_start_high[FS_VCO4_INDEX];
        cc112xSpiWriteReg(CC112X_FS_VCO4, &writeByte, 1);
        writeByte = calResults_for_vcdac_start_high[FS_CHP_INDEX];
        cc112xSpiWriteReg(CC112X_FS_CHP, &writeByte, 1);
    }
    else 
    {
        writeByte = calResults_for_vcdac_start_mid[FS_VCO2_INDEX];
        cc112xSpiWriteReg(CC112X_FS_VCO2, &writeByte, 1);
        writeByte = calResults_for_vcdac_start_mid[FS_VCO4_INDEX];
        cc112xSpiWriteReg(CC112X_FS_VCO4, &writeByte, 1);
        writeByte = calResults_for_vcdac_start_mid[FS_CHP_INDEX];
        cc112xSpiWriteReg(CC112X_FS_CHP, &writeByte, 1);
    }
}


ttc_radio_errorcode_e radio_cc1120_init_SmartRF(ttc_radio_generic_t* Radio_Generic) {
    radio_cc1120_driver_t* DriverCfg = radio_cc1120_get_config(Radio_Generic); // checks pointer too

    // apply radio configuration
    ttc_radio_errorcode_e Error = tre_OK;

    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_IOCFG3         , SMARTRF_SETTING_0x0f); //0x0000,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_IOCFG2         , SMARTRF_SETTING_0x0f); //0x0001,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_IOCFG1         , SMARTRF_SETTING_0x0f); //0x0002,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_IOCFG0         , SMARTRF_SETTING_0x0f); //0x0003,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_SYNC3          , SMARTRF_SETTING_0x0f); //0x0004,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_SYNC2          , SMARTRF_SETTING_0x0f); //0x0005,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_SYNC1          , SMARTRF_SETTING_0x0f); //0x0006,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_SYNC0          , SMARTRF_SETTING_0x0f); //0x0007,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_SYNC_CFG1      , SMARTRF_SETTING_0x0f); //0x0008,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_SYNC_CFG0      , SMARTRF_SETTING_0x0f); //0x0009,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_DEVIATION_M    , SMARTRF_SETTING_0x0f); //0x000A,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_MODCFG_DEV_E   , SMARTRF_SETTING_0x0f); //0x000B,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_DCFILT_CFG     , SMARTRF_SETTING_0x0f); //0x000C,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_PREAMBLE_CFG1  , SMARTRF_SETTING_0x0f); //0x000D,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_PREAMBLE_CFG0  , SMARTRF_SETTING_0x0f); //0x000E,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_FREQ_IF_CFG    , SMARTRF_SETTING_0x0f); //0x000F,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_IQIC           , SMARTRF_SETTING_0x0f); //0x0010,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_CHAN_BW        , SMARTRF_SETTING_0x0f); //0x0011,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_MDMCFG1        , SMARTRF_SETTING_0x0f); //0x0012,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_MDMCFG0        , SMARTRF_SETTING_0x0f); //0x0013,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_DRATE2         , SMARTRF_SETTING_0x0f); //0x0014,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_DRATE1         , SMARTRF_SETTING_0x0f); //0x0015,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_DRATE0         , SMARTRF_SETTING_0x0f); //0x0016,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_AGC_REF        , SMARTRF_SETTING_0x0f); //0x0017,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_AGC_CS_THR     , SMARTRF_SETTING_0x0f); //0x0018,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_AGC_GAIN_ADJUST, SMARTRF_SETTING_0x0f); //0x0019,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_AGC_CFG3       , SMARTRF_SETTING_0x0f); //0x001A,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_AGC_CFG2       , SMARTRF_SETTING_0x0f); //0x001B,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_AGC_CFG1       , SMARTRF_SETTING_0x0f); //0x001C,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_AGC_CFG0       , SMARTRF_SETTING_0x0f); //0x001D,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_FIFO_CFG       , SMARTRF_SETTING_0x0f); //0x001E,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_DEV_ADDR       , SMARTRF_SETTING_0x0f); //0x001F,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_SETTLING_CFG   , SMARTRF_SETTING_0x0f); //0x0020,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_FS_CFG         , SMARTRF_SETTING_0x0f); //0x0021,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_WOR_CFG1       , SMARTRF_SETTING_0x0f); //0x0022,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_WOR_CFG0       , SMARTRF_SETTING_0x0f); //0x0023,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_WOR_EVENT0_MSB , SMARTRF_SETTING_0x0f); //0x0024,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_WOR_EVENT0_LSB , SMARTRF_SETTING_0x0f); //0x0025,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_PKT_CFG2       , SMARTRF_SETTING_0x0f); //0x0026,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_PKT_CFG1       , SMARTRF_SETTING_0x0f); //0x0027,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_PKT_CFG0       , SMARTRF_SETTING_0x0f); //0x0028,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_RFEND_CFG1     , SMARTRF_SETTING_0x0f); //0x0029,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_RFEND_CFG0     , SMARTRF_SETTING_0x0f); //0x002A,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_PA_CFG2        , SMARTRF_SETTING_0x0f); //0x002B,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_PA_CFG1        , SMARTRF_SETTING_0x0f); //0x002C,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_PA_CFG0        , SMARTRF_SETTING_0x0f); //0x002D,
    if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_CC112X_PKT_LEN        , SMARTRF_SETTING_0x0f); //0x002E,


    u8_t Compare = 0;
    if (!Error) Error = radio_cc1120_spi_register_access(DriverCfg, rcr_ADDR, &Compare, cam_single_read, 1); //  0x01; // Device address.
    Assert_Radio(Compare == SMARTRF_SETTING_ADDR, ec_DeviceNotFound); // written value could not be verified

    return Error;
}
ttc_radio_errorcode_e radio_cc1120_packet_send(ttc_radio_generic_t* Radio_Generic, u8_t TargetAddress, u8_t Size, u8_t* Data) {
    radio_cc1120_driver_t* DriverCfg = radio_cc1120_get_config(Radio_Generic); // checks pointer too
    Assert_Radio(Size, ec_InvalidArgument);
    Assert_Radio(Data, ec_InvalidArgument);

    ttc_radio_errorcode_e Error = tre_OK;
    Base_t StartTime =0;
    u16_t TimeTxPreparation = 0 ;

    Assert_Radio(Size, ec_InvalidArgument);
    Assert_Radio(Size <= RADIO_cc1120_PAYLOAD_SIZE, ec_InvalidArgument);

    LockTransaction(DriverCfg->SPI_Index, radio_cc1120_packet_send);

    if(1){ //D //disable complete radio transmission
        if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_TXFIFO, Size + 1);
        Assert_Radio(Error == tre_OK, ec_UNKNOWN);
        if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_TXFIFO, TargetAddress);
        Assert_Radio(Error == tre_OK, ec_UNKNOWN);
        if (!Error) Error = radio_cc1120_spi_register_access(DriverCfg, rcr_TXFIFO, Data, cam_burst_write, Size);
        Assert_Radio(Error == tre_OK, ec_UNKNOWN);

        //ttc_task_begin_criticalsection();

        if (!Error) Error = radio_cc1120_spi_strobe(DriverCfg, rcr_STX);

        u8_t TransmitState = 0;
        StartTime = ttc_task_get_tick_count();
        while ( (TransmitState = _radio_cc1120_get_state(Radio_Generic)) != rcm_TX) { //wait for radio to reach TX State
            switch(TransmitState){
            case rcm_TXFIFO_UNDERFLOW:
                return _radio_cc1120_handle_error(Radio_Generic, tre_TxFIFO_Underrun, TRUE);
            case rcm_RXFIFO_OVERFLOW:
                return _radio_cc1120_handle_error(Radio_Generic, tre_RxFIFO_Overrun, TRUE);
            case rcm_IDLE:
                return _radio_cc1120_handle_error(Radio_Generic, tre_UnexpectedState, TRUE);
            case rcm_RX:
                return _radio_cc1120_handle_error(Radio_Generic, tre_UnexpectedState, TRUE);
            default :
                break;

            }

            if((ttc_task_get_tick_count() - StartTime)>RADIO_cc1120_TRANSMIT_TIMEOUT)
                return _radio_cc1120_handle_error(Radio_Generic, tre_TX_Timeout, TRUE);

            Radio_Generic->TaskYield();
        }
        TimeTxPreparation = ttc_task_get_tick_count() - StartTime;

        while ( (TransmitState = _radio_cc1120_get_state(Radio_Generic)) != rcm_RX) { //wait for radio to reach RX State ttc_task_get_tick_count() always returns the same value: 11130
            switch(TransmitState){
            case rcm_TXFIFO_UNDERFLOW:
                return _radio_cc1120_handle_error(Radio_Generic, tre_TxFIFO_Underrun, TRUE);
            case rcm_RXFIFO_OVERFLOW:
                return _radio_cc1120_handle_error(Radio_Generic, tre_RxFIFO_Overrun, TRUE);
            case rcm_IDLE:
                return _radio_cc1120_handle_error(Radio_Generic, tre_UnexpectedState, TRUE);
            default :
                break;
            }

            Radio_Generic->TaskYield();
        }
        //ttc_task_end_criticalsection();
#ifdef TTC_DEBUG_RADIO
        u16_t TimeTxMode = ttc_task_get_tick_count() - StartTime - TimeTxPreparation;

        if (Radio_Generic->Max_TimeTxMode < TimeTxMode)
            Radio_Generic->Max_TimeTxMode = TimeTxMode;
        if (Radio_Generic->Max_TimeTxPreparation < TimeTxPreparation)
            Radio_Generic->Max_TimeTxPreparation = TimeTxPreparation;
#endif
    } //disable complete radio transmission

    UnlockTransaction(DriverCfg->SPI_Index);

    return Error;
}
                 u8_t radio_cc1120_packet_receive(ttc_radio_generic_t* Radio_Generic, u8_t* Buffer, u8_t MaxSize, u8_t* TargetAdress, u8_t* RSSI) {
    radio_cc1120_driver_t* DriverCfg = radio_cc1120_get_config(Radio_Generic); // checks pointer too
    ttc_radio_errorcode_e Error = tre_OK;
    Assert_Radio(RADIO_cc1120_PAYLOAD_SIZE <= MaxSize, ec_InvalidArgument); // Buffer[] is too small for this packet

    LockTransaction(DriverCfg->SPI_Index, radio_cc1120_packet_receive);

    u8_t RX_BytesPending = 0 ;
    if (!Error) Error = radio_cc1120_spi_register_access(DriverCfg, rcr_RXBYTES, &RX_BytesPending, cam_single_read, 1);
    u8_t PacketSize = 0;

    if (RX_BytesPending) {
        RX_BytesPending = RX_BytesPending& 0x7F; //RX_BytesPending contains MSB that indicates if RXFIFO is overflowed
        if (RX_BytesPending > cc1120_SIZE_RX_FIFO) { // invalid size: reread value via SPI
            //get bytes again, if a value > cc1120_SIZE_RX_FIFO is read from the RX Bytes Register.
            if (!Error) Error = radio_cc1120_spi_register_access(DriverCfg, rcr_RXBYTES, &RX_BytesPending, cam_single_read, 1); //D
            RX_BytesPending = RX_BytesPending& 0x7F;
        }
        if ( (RX_BytesPending > cc1120_SIZE_RX_FIFO) ||
             (RX_BytesPending > TTC_NETWORK_MAX_PAYLOAD_SIZE)
        ) { // invalid packet size: flush rx-buffer
            _radio_cc1120_handle_error(Radio_Generic, tre_RxFIFO_Overrun, FALSE);
        }
        Amount_NonEmptyPackets++; //D
        if (!Error) Error = radio_cc1120_spi_register_access(DriverCfg, rcr_RXFIFO, &PacketSize,  cam_single_read, 1);
        if (!Error) Error = radio_cc1120_spi_register_access(DriverCfg, rcr_RXFIFO, TargetAdress, cam_single_read, 1);
        PacketSize--; //remove Target Address


        if (PacketSize > RADIO_cc1120_PAYLOAD_SIZE) { // invalid size: flush fifo
            PacketSize = 0;
            Error = tre_RxCorruptPacket;
        }

        u8_t Bytes2Read = PacketSize;
        while (Bytes2Read--) {
            if (!Error) *Buffer++ = radio_cc1120_spi_register_single_read(DriverCfg, rcr_RXFIFO);
        }
    }

    if (Error != tre_OK) {
        PacketSize = 0;
        _radio_cc1120_handle_error(Radio_Generic, Error, TRUE); // unlocks transaction
    }
    else
        UnlockTransaction(DriverCfg->SPI_Index);

    return PacketSize;
}
ttc_radio_errorcode_e radio_cc1120_set_device_address(ttc_radio_generic_t* Radio_Generic, u8_t NewDeviceAddress) {
    radio_cc1120_driver_t* DriverCfg = radio_cc1120_get_config(Radio_Generic); // checks pointer too

    ttc_radio_errorcode_e Error = tre_OK;
    if (Radio_Generic->DeviceAddress != NewDeviceAddress)
    {
        LockTransaction(DriverCfg->SPI_Index, radio_cc1120_packet_receive);
        if (!Error) Error = radio_cc1120_spi_register_access(DriverCfg, rcr_ADDR, &NewDeviceAddress, cam_single_write, 1 );
        Radio_Generic->DeviceAddress = NewDeviceAddress;
        UnlockTransaction(DriverCfg->SPI_Index);
    }
    return Error;
}
ttc_radio_errorcode_e radio_cc1120_set_channel(ttc_radio_generic_t* Radio_Generic, u32_t NewChannelIndex) {
    radio_cc1120_driver_t* DriverCfg = radio_cc1120_get_config(Radio_Generic); // checks pointer too
    ttc_radio_errorcode_e Error = tre_OK;

    if (Radio_Generic->ChannelTx != NewChannelIndex)
    {
        LockTransaction(DriverCfg->SPI_Index, radio_cc1120_packet_receive);
        u8_t NewChannelIndex8 = (u8_t) 0xff & NewChannelIndex;
        if (!Error) Error = radio_cc1120_spi_register_access(DriverCfg, rcr_CHANNR, &NewChannelIndex8, cam_single_write, 1);
        if (!Error) {
            Radio_Generic->ChannelTx = NewChannelIndex;
            Radio_Generic->ChannelRx = NewChannelIndex;
        }
        UnlockTransaction(DriverCfg->SPI_Index);
    }
    return Error;
}
ttc_radio_errorcode_e radio_cc1120_set_power_tx(ttc_radio_generic_t* Radio_Generic, s8_t Level) {
    radio_cc1120_driver_t* DriverCfg = radio_cc1120_get_config(Radio_Generic); // checks pointer too

    LockTransaction(DriverCfg->SPI_Index, radio_cc1120_set_power_tx);

    // determine power-setting for 868MHz according to cc1120-datasheet p.59
    u8_t PowerSetting = 0;
    if (Level > 11) {
        Level = 10;
        PowerSetting = 0xc0;
    }
    else if (Level > 8) {
        Level = 10;
        PowerSetting = 0xc5;
    }
    else if (Level > 6) {
        Level = 7;
        PowerSetting = 0xcd;
    }
    else if (Level > 2) {
        Level = 5;
        PowerSetting = 0x86;
    }
    else if (Level > -3) {
        Level = 0;
        PowerSetting = 0x50;
    }
    else if (Level > -8) {
        Level = -6;
        PowerSetting = 0x37;
    }
    else if (Level > -13) {
        Level = -10;
        PowerSetting = 0x26;
    }
    else if (Level > -17) {
        Level = -15;
        PowerSetting = 0x1d;
    }
    else if (Level > -25) {
        Level = -20;
        PowerSetting = 0x17;
    }
    else {
        Level = -30;
        PowerSetting = 0x03;
    }

    Radio_Generic->LevelTx = Level;
    u8_t PA_Table[8] = { 0,0,0,0,0,0,0,0 };
    PA_Table[0] = PowerSetting;
    PA_Table[1] = PowerSetting;
    ttc_radio_errorcode_e Error = radio_cc1120_spi_register_access(DriverCfg, rcr_PATABLE, PA_Table, cam_burst_write, 8);

    UnlockTransaction(DriverCfg->SPI_Index);

    return Error;
}
ttc_radio_errorcode_e radio_cc1120_set_operating_mode(ttc_radio_generic_t* Radio_Generic) {
    radio_cc1120_driver_t* DriverCfg = radio_cc1120_get_config(Radio_Generic); // checks pointer too
    ttc_radio_errorcode_e Error = tre_OK;

    LockTransaction(DriverCfg->SPI_Index, radio_cc1120_set_operating_mode);
    radio_cc1120_marcstate_e CurrentMode = radio_cc1120_spi_register_single_read(DriverCfg, rcr_MARCSTATE);

    if (CurrentMode == rcm_RXFIFO_OVERFLOW) {
        UnlockTransaction(DriverCfg->SPI_Index); // _radio_cc1120_handle_error() might call us recursively!
        _radio_cc1120_handle_error(Radio_Generic, tre_RxFIFO_Overrun, FALSE);
        LockTransaction(DriverCfg->SPI_Index, radio_cc1120_set_operating_mode);
    }
    if (CurrentMode == rcm_TXFIFO_UNDERFLOW) {
        UnlockTransaction(DriverCfg->SPI_Index); // _radio_cc1120_handle_error() might call us recursively!
        _radio_cc1120_handle_error(Radio_Generic, tre_TxFIFO_Underrun, FALSE);
        LockTransaction(DriverCfg->SPI_Index, radio_cc1120_set_operating_mode);
    }

    switch (Radio_Generic->Mode) {
    case trm_Idle: {
        radio_cc1120_spi_strobe(DriverCfg, rcr_SIDLE);
        break;
    }
    case trm_Reset: {
        radio_cc1120_spi_strobe(DriverCfg, rcr_SRES);
        break;
    }
    case trm_PowerOff:
    case trm_Sleep: {
        radio_cc1120_spi_strobe(DriverCfg, rcr_SPWD);
        break;
    }
    case trm_PowerOn: {
        if (!Error) Error = radio_cc1120_spi_strobe(DriverCfg, rcr_SNOP);
        if (!Error) Error = radio_cc1120_spi_strobe(DriverCfg, rcr_SNOP);
        if (!Error) Error = radio_cc1120_spi_strobe(DriverCfg, rcr_SNOP);

        // reload configuration
        if (!Error) Error = radio_cc1120_init(Radio_Generic);
    }
    case trm_Transmit: {
        radio_cc1120_spi_strobe(DriverCfg, rcr_STX);
        break;
    }
    case trm_Receive: {
        radio_cc1120_spi_strobe(DriverCfg, rcr_SRX);
        break;
    }
    case trm_TransmitReceive: {
    default:
            Assert_Radio(0, ec_InvalidArgument); // not supported by this radio
            break;
        }
    }

    UnlockTransaction(DriverCfg->SPI_Index);
    Base_t TimeOut = 10000;
    while (radio_cc1120_check_mode(Radio_Generic) && (TimeOut > 0)) {
        TimeOut--;
        ttc_task_yield();
    }
    if (TimeOut == 0)
        Error = tre_TimeOut;

    return Error;
}
                  BOOL radio_cc1120_check_mode(ttc_radio_generic_t* Radio_Generic) {
    radio_cc1120_driver_t* DriverCfg = radio_cc1120_get_config(Radio_Generic); // checks pointer too

    LockTransaction(DriverCfg->SPI_Index, radio_cc1120_set_operating_mode);
    BOOL IllegalMode = FALSE;

    switch (Radio_Generic->Mode) {
    case trm_Idle: {
        if (radio_cc1120_spi_register_single_read(DriverCfg, rcr_MARCSTATE) != rcm_IDLE)
            IllegalMode = TRUE;
        break;
    }
    case trm_PowerOn: {
        radio_cc1120_marcstate_e CurrentMode = radio_cc1120_spi_register_single_read(DriverCfg, rcr_MARCSTATE);
        if ( (CurrentMode != rcm_IDLE) && (CurrentMode != rcm_TX) && (CurrentMode != rcm_RX) )
             IllegalMode = TRUE;
    }
    case trm_Transmit: {
        radio_cc1120_marcstate_e CurrentMode = radio_cc1120_spi_register_single_read(DriverCfg, rcr_MARCSTATE);
        if (CurrentMode != rcm_TX)
             IllegalMode = TRUE;
        break;
    }
    case trm_Receive: {
        radio_cc1120_marcstate_e CurrentMode = radio_cc1120_spi_register_single_read(DriverCfg, rcr_MARCSTATE);
        if (CurrentMode != rcm_RX)
             IllegalMode = TRUE;
        break;
    }
    default: break;
    }
    UnlockTransaction(DriverCfg->SPI_Index);

    return IllegalMode;
}
                  u8_t radio_cc1120_get_spi_index(u8_t RadioIndex) {
    Assert_Radio(RadioIndex > 0, ec_InvalidArgument); // logical index starts at 1

    switch (RadioIndex) {
#ifdef TTC_RADIO1_SPI_INDEX
    case 1: return TTC_RADIO1_SPI_INDEX;
#endif
#ifdef TTC_RADIO2_SPI_INDEX
    case 2: return TTC_RADIO2_SPI_INDEX;
#endif
#ifdef TTC_RADIO3_SPI_INDEX
    case 3: return TTC_RADIO3_SPI_INDEX;
#endif
#ifdef TTC_RADIO4_SPI_INDEX
    case 4: return TTC_RADIO4_SPI_INDEX;
#endif
#ifdef TTC_RADIO5_SPI_INDEX
    case 5: return TTC_RADIO5_SPI_INDEX;
#endif
    default: break;
    }
    Assert_Radio(0, ec_InvalidConfiguration); // indexed radio is not activated!
    return 0;
}
                  u8_t ISR_Call_DEBUG_VALUE = 0; //D
radio_cc1120_driver_t* radio_cc1120_get_config(ttc_radio_generic_t* Radio_Generic) {
    Assert_Radio(Radio_Generic, ec_InvalidArgument);

    radio_cc1120_driver_t* DriverCfg = (radio_cc1120_driver_t*) Radio_Generic->DriverCfg;
    if (!DriverCfg) {
        DriverCfg = (radio_cc1120_driver_t*) ttc_memory_alloc_zeroed( sizeof(radio_cc1120_driver_t) );
        Radio_Generic->DriverCfg = (ttc_radio_driver_t*) DriverCfg;
        DriverCfg->SPI_Index = radio_cc1120_get_spi_index(Radio_Generic->LogicalIndex);

        // ToDo: SP - fill out default driver configuration
        DriverCfg->RadioIndex = Radio_Generic->LogicalIndex;
        DriverCfg->Lock = ttc_mutex_create();
    }
    return DriverCfg;
}
                 void _radio_cc1120_isr_receive(physical_index_t PhysicalIndex, void* Argument) {
    (void) PhysicalIndex;
    Assert_Radio(Argument, ec_InvalidArgument);
#ifdef TTC_DEBUG_RADIO
    //ttc_gpio_set(TTC_DEBUG_RADIO_PIN2);
    ISR_Call_DEBUG_VALUE = 1 - ISR_Call_DEBUG_VALUE; //D
    //? ttc_gpio_setn(TTC_DEBUG_RADIO_PIN2,ISR_Call_DEBUG_VALUE);//D
#endif
    ttc_radio_generic_t* Radio_Generic = (ttc_radio_generic_t*) Argument;
    if (Radio_Generic->MutexPacketReceived)
        ttc_mutex_unlock(Radio_Generic->MutexPacketReceived);

#ifdef TTC_DEBUG_RADIO
    //ttc_gpio_clr(TTC_DEBUG_RADIO_PIN2);
#endif
}
                 u8_t _radio_cc1120_get_state(ttc_radio_generic_t* Radio_Generic) {
    Assert_Radio(Radio_Generic, ec_InvalidArgument);
    radio_cc1120_driver_t* DriverCfg = (radio_cc1120_driver_t*) Radio_Generic->DriverCfg;

    u8_t MarcState;
    ttc_radio_errorcode_e Error = radio_cc1120_spi_register_access(DriverCfg,
                                                               if (!Error) Error = radio_cc1120_spi_register_single_write(DriverCfg, rcr_MARCSTATE,
                                                                   &MarcState,
                                                                   cam_single_read,
                                                                   1
                                                                   );
    Assert_Radio(Error == tre_OK, ec_UNKNOWN);

    return MarcState;
}
ttc_radio_errorcode_e _radio_cc1120_handle_error(ttc_radio_generic_t* Radio_Generic, ttc_radio_errorcode_e Error, BOOL UnlockTransaction) {
    radio_cc1120_driver_t* DriverCfg = radio_cc1120_get_config(Radio_Generic); // checks pointer too

    ttc_radio_errorcode_e Error_return = tre_OK;

#ifdef TTC_DEBUG_RADIO
    Radio_Generic->Amount_Error_Handler_Calls++; //D
    Assert_Radio(Radio_Generic->Amount_Error_Handler_Calls<50, ec_Debug);//D
#endif
    if (Error == tre_TX_Timeout){
#ifdef TTC_DEBUG_RADIO
        Radio_Generic->Amount_Error_TX_Timeout++; //D
        Assert_Radio(Radio_Generic->Amount_Error_TX_Timeout<50, ec_Debug);//D
#endif
        if(_radio_cc1120_get_state(Radio_Generic) == rcm_RX){ // assume that we missed the TX State because radio is allready back in RX State
#ifdef TTC_DEBUG_RADIO
            Radio_Generic->Amount_Error_Missed_TXStates++; //D
#endif
            Error_return = tre_OK;
        }
        else {
            Assert_Radio(FALSE, ec_UNKNOWN);
            //assert if state is != rcm_RX because then we have an unexpected state
            //this assert could be replaced by another  _radio_cc1120_handle_error call with tre_UnexpectedState
            //this requires recursive locks.
        }
    } else if(Error == tre_RxCorruptPacket){
#ifdef TTC_DEBUG_RADIO
        Radio_Generic->Amount_Error_RxCorruptPacket++; //D
#endif
        radio_cc1120_spi_strobe(DriverCfg, rcr_SFRX);
        radio_cc1120_spi_strobe(DriverCfg, rcr_SRX);
    }
    else if(Error == tre_UnexpectedState){
#ifdef TTC_DEBUG_RADIO
        Radio_Generic->Amount_Error_UnexpectedState++; //D
#endif
        u8_t CurrentState = _radio_cc1120_get_state(Radio_Generic) ;
        if(CurrentState == rcm_TXFIFO_UNDERFLOW){
#ifdef TTC_DEBUG_RADIO
            Radio_Generic->Amount_Error_TXFIFO_UNDERFLOW++; //D
#endif
            radio_cc1120_spi_strobe(DriverCfg, rcr_SFTX);
        }
        else if(CurrentState == rcm_RXFIFO_OVERFLOW){
#ifdef TTC_DEBUG_RADIO
            Radio_Generic->Amount_Error_RXFIFO_OVERFLOW++; //D
#endif
            radio_cc1120_spi_strobe(DriverCfg, rcr_SFRX);
        }
        //ToDo: needs to bring back the radio to an known state.
    }
    else if(Error==tre_RxFIFO_Overrun){
        radio_cc1120_spi_strobe(DriverCfg, rcr_SFRX);
        radio_cc1120_spi_strobe(DriverCfg, rcr_SIDLE);
        radio_cc1120_spi_strobe(DriverCfg, rcr_SRX);
#ifdef TTC_DEBUG_RADIO
        Radio_Generic->Amount_Error_RXFIFO_OVERFLOW++; //D
#endif
    }
    else {
        radio_cc1120_spi_strobe(DriverCfg, rcr_SFTX);
        radio_cc1120_spi_strobe(DriverCfg, rcr_SFRX);
        radio_cc1120_spi_strobe(DriverCfg, rcr_SIDLE);
        if (!Error_return) Error_return = radio_cc1120_init(Radio_Generic);
        radio_cc1120_spi_strobe(DriverCfg, rcr_SRX);
    }
    if (Error_return)
        Assert_Halt_EC(ec_UNKNOWN); // Break here


    UnlockTransaction(DriverCfg->SPI_Index);

    ttc_radio_set_operating_mode(Radio_Generic->LogicalIndex, Radio_Generic->Mode); //switch radio back to its configured Mode

    if (UnlockTransaction){
        UnlockTransaction(DriverCfg->SPI_Index);
    }
    else {
        LockTransaction(DriverCfg->SPI_Index,_radio_cc1120_handle_error);
    }

//    radio_cc1120_spi_strobe(DriverCfg, rcr_SFRX);
//    radio_cc1120_spi_strobe(DriverCfg, rcr_SIDLE);
//    radio_cc1120_spi_strobe(DriverCfg, rcr_SRX);
    return Error_return;
}

//} Implemented Functions
