/*{ stm32_i2c::.c ************************************************
 
                      The ToolChain
                      
   Universal Synchronous Asynchronous Receiver and Transmitters (stm32_I2Cs)
   for ARM CortexM3 microcontrollers.
   Currently tested to comply to STM32F10x types.
   
   written by Gregor Rebel 2012
   
   Note: See ttc_i2c.h for description of architecture independent I2C implementation.
 
}*/

#define USE_OWN_I2C_CODE 1  // do not use buggy I2C-code of StdPeripheral whenever possible

#include "stm32_i2c.h"

// stores configuration data of every initialized I2C-interface
stm32_i2c_architecture_t* stm32_I2C_Configs[TTC_AMOUNT_I2CS];


u8_t si2c_Initialized=0;  // ==1: stm32_I2C_Configs has been initialized

//{ Function definitions ********************************************************

void stm32_i2c_reset_all() {
  memset(stm32_I2C_Configs, 0, sizeof(stm32_i2c_architecture_t*) * TTC_AMOUNT_I2CS);
  si2c_Initialized=1;
}
ttc_i2c_errorcode_e stm32_i2c_get_defaults(u8_t I2C_Index, ttc_i2c_generic_t* I2C_Generic) {
    if (!si2c_Initialized) stm32_i2c_reset_all();
#ifndef TTC_NO_ARGUMENT_CHECKS //{

    Assert(I2C_Index > 0,    ec_InvalidArgument);
    Assert(I2C_Generic != NULL, ec_NULL);
    if (I2C_Index > TTC_AMOUNT_I2CS) return tie_DeviceNotFound;
#endif //}

    // assumption: *I2C_Generic has been zeroed
    I2C_Generic->Flags.All                   = 0;
    I2C_Generic->Flags.Bits.Master           = 1;
    I2C_Generic->Flags.Bits.ModeFast         = 0;
    I2C_Generic->Flags.Bits.Slave            = 0;
    I2C_Generic->Flags.Bits.DMA_TX           = 0;
    I2C_Generic->Flags.Bits.DMA_RX           = 0;
    I2C_Generic->Flags.Bits.SMBus            = 0;
    I2C_Generic->Flags.Bits.SMBus_Alert      = 0;
    I2C_Generic->Flags.Bits.SMBus_Host       = 0;
    I2C_Generic->Flags.Bits.SMBus_ARP        = 0;
    I2C_Generic->Flags.Bits.DutyCycle_16_9   = 0;
    I2C_Generic->Flags.Bits.Acknowledgement  = 1;
    I2C_Generic->Flags.Bits.Adress_10Bit     = 0;
    I2C_Generic->Flags.Bits.PacketErrorCheck = 0;
    I2C_Generic->Flags.Bits.Slave_NoStretch  = 0;
    I2C_Generic->Flags.Bits.GeneralCalls     = 0;
    I2C_Generic->Flags.Bits.Interrupt_Events = 0;
    I2C_Generic->Flags.Bits.Interrupt_Errors = 0;
    I2C_Generic->ClockSpeed                  = 10000; // Standard Mode

    I2C_Generic->OwnAddress = 1;
    I2C_Generic->TimeOut    = 200000;

    return tie_OK;
}
ttc_i2c_errorcode_e stm32_i2c_get_features(u8_t I2C_Index, ttc_i2c_generic_t* I2C_Generic) {
    if (!si2c_Initialized) stm32_i2c_reset_all();
#ifndef TTC_NO_ARGUMENT_CHECKS //{
    // assumption: *I2C_Generic has been zeroed

    Assert(I2C_Index > 0,    ec_InvalidArgument);
    Assert(I2C_Generic != NULL, ec_NULL);
    if (I2C_Index > TTC_AMOUNT_I2CS) return tie_DeviceNotFound;
#endif //}

    I2C_Generic->Flags.All                   = 0;
    I2C_Generic->Flags.Bits.Master           = 1;
    I2C_Generic->Flags.Bits.ModeFast         = 1;
    I2C_Generic->Flags.Bits.Slave            = 0; // ToDo: implement
    I2C_Generic->Flags.Bits.DMA_TX           = 0; // ToDo: implement
    I2C_Generic->Flags.Bits.DMA_RX           = 0; // ToDo: implement
    I2C_Generic->Flags.Bits.SMBus            = 0; // ToDo: implement
    I2C_Generic->Flags.Bits.SMBus_Alert      = 0; // ToDo: implement 
    I2C_Generic->Flags.Bits.SMBus_Host       = 0; // ToDo: implement
    I2C_Generic->Flags.Bits.SMBus_ARP        = 0; // ToDo: implement
    I2C_Generic->Flags.Bits.DutyCycle_16_9   = 1;
    I2C_Generic->Flags.Bits.Acknowledgement  = 1;
    I2C_Generic->Flags.Bits.Adress_10Bit     = 0; // ToDo: implement
    I2C_Generic->Flags.Bits.PacketErrorCheck = 1;
    I2C_Generic->Flags.Bits.Slave_NoStretch  = 0; // ToDo: implement
    I2C_Generic->Flags.Bits.GeneralCalls     = 1;
    I2C_Generic->Flags.Bits.Interrupt_Events = 0; // ToDo: implement
    I2C_Generic->Flags.Bits.Interrupt_Errors = 0; // ToDo: implement
    I2C_Generic->ClockSpeed                  = 400000;
    I2C_Generic->TimeOut                     = -1; // max positive value

    I2C_TypeDef* MyI2C = NULL;
    switch (I2C_Index) {           // find corresponding I2C as defined by makefile.100_board_*
#ifdef TTC_I2C1
      case 1: MyI2C = TTC_I2C1; break;
#endif
#ifdef TTC_I2C2
      case 2: MyI2C = TTC_I2C2; break;
#endif
      default: Assert(0, ec_UNKNOWN); break; // No TTC_I2Cn defined! Check your makefile.100_board_* file!
    }
    switch ( (u32_t) MyI2C ) {  // find amount of available remapping-layouts (-> RM0008 p. 176)
    case (u32_t) I2C1:
        I2C_Generic->Layout=1;     // I2C1 provides 1 alternate pin layout
        break;
    case (u32_t) I2C2:
        I2C_Generic->Layout=0;     // I2C2 provides 0 alternate pin layouts
        break;
    default: Assert(0, ec_UNKNOWN);
    }

    switch (I2C_Index) {           // disable features requiring unconfigured pins
      case 1: {
          #ifndef TTC_I2C1_SMBAL
            I2C_Generic->Flags.Bits.SMBus_Alert = 0;
          #endif
          
          break;
      }
      case 2: {
          #ifndef TTC_I2C2_SMBAL
            I2C_Generic->Flags.Bits.SMBus_Alert = 0;
          #endif
          
          break;
      }
      default: Assert(0, ec_UNKNOWN); return tie_DeviceNotFound;; // I2C misconfigured! Check your makefile.100_board_* file!
    }
    
    return tie_OK;
}
ttc_i2c_errorcode_e stm32_i2c_init(u8_t I2C_Index, ttc_i2c_generic_t* I2C_Generic, stm32_i2c_architecture_t* I2C_Arch) {
    if (!si2c_Initialized) stm32_i2c_reset_all();
#ifndef TTC_NO_ARGUMENT_CHECKS
    // assumption: *I2C_Arch has been zeroed

    Assert(I2C_Index > 0, ec_InvalidArgument);
    Assert(I2C_Generic != NULL, ec_NULL) ;
#endif

    if (1) { // validate I2C_Features
        ttc_i2c_generic_t I2C_Features;
        ttc_i2c_errorcode_e Error=stm32_i2c_get_features(I2C_Index, &I2C_Features);
        if (Error) return Error;
        I2C_Generic->Flags.All &= I2C_Features.Flags.All; // mask out unavailable flags
        if (I2C_Generic->Layout > I2C_Features.Layout)
            I2C_Generic->Layout = I2C_Features.Layout;
        if (I2C_Generic->ClockSpeed > I2C_Features.ClockSpeed)
            I2C_Generic->ClockSpeed = I2C_Features.ClockSpeed;
        if (I2C_Generic->Flags.Bits.Adress_10Bit)
            I2C_Generic->OwnAddress &= 0x03ff; // create 10 bit TargetAddress
        else 
            I2C_Generic->OwnAddress &= 0x007f; // create  7 bit TargetAddress

        if (I2C_Features.Flags.Bits.Master)     // override slave setting
            I2C_Features.Flags.Bits.Slave = 0;
        if (I2C_Features.Flags.Bits.Slave == 0) // not slave => being master
            I2C_Features.Flags.Bits.Master = 1;
    }
    switch (I2C_Index) {                // find I2C corresponding to I2C_index as defined by makefile.100_board_*
#ifdef TTC_I2C1
      case 1: I2C_Arch->Base = (I2C_t*) TTC_I2C1; break;
#endif
#ifdef TTC_I2C2
      case 2: I2C_Arch->Base = (I2C_t*) TTC_I2C2; break;
#endif
      default: Assert(0, ec_UNKNOWN); break; // No TTC_I2Cn defined! Check your makefile.100_board_* file!
    }
    I2C_Generic->Layout = _stm32_i2c_get_pins(I2C_Index, I2C_Arch);
    if (1) { // compare chosen pin layout to configured pins
        //tie_InvalidPinConfig
        stm32_Port_t Port;
        switch (I2C_Index) { // check if board has defined the TX-pin of I2C to init
          
          case 1: { // check configuration of I2C #1
  #ifdef TTC_I2C1_SDA
              stm32_gpio_variable(&Port, TTC_I2C1_SDA);
              Assert(Port.GPIOx == I2C_Arch->PortSDA.GPIOx, ec_InvalidArgument); // ERROR: configured GPIO cannot be used in current remapping layout!
              Assert(Port.Pin   == I2C_Arch->PortSDA.Pin,   ec_InvalidArgument); // ERROR: configured Pin  cannot be used in current remapping layout!
  #else
              I2C_Arch->PortSDA.GPIOx = NULL;
  #endif
  #ifdef TTC_I2C1_SCL
              stm32_gpio_variable(&Port, TTC_I2C1_SCL);
              Assert(Port.GPIOx == I2C_Arch->PortSCL.GPIOx, ec_InvalidArgument); // ERROR: configured GPIO cannot be used in current remapping layout!
              Assert(Port.Pin   == I2C_Arch->PortSCL.Pin,   ec_InvalidArgument); // ERROR: configured Pin  cannot be used in current remapping layout!
  #else
              I2C_Arch->PortSCL.GPIOx = NULL;
  #endif
#ifdef TTC_I2C1_SMBAL
              if (I2C_Generic->Flags.Bits.SMBus_Alert) {
                  stm32_gpio_variable(&Port, TTC_I2C1_SMBAL);
                  Assert(Port.GPIOx == I2C_Arch->PortSMBAL.GPIOx, ec_InvalidArgument); // ERROR: configured GPIO cannot be used in current remapping layout!
                  Assert(Port.Pin   == I2C_Arch->PortSMBAL.Pin,   ec_InvalidArgument); // ERROR: configured Pin  cannot be used in current remapping layout!
              }
#else
              I2C_Arch->PortSMBAL.GPIOx = NULL;
  #endif
              break;
          }
          case 2: { // check configuration of I2C #2
  #ifdef TTC_I2C2_SDA
              stm32_gpio_variable(&Port, TTC_I2C2_SDA);
              Assert(Port.GPIOx == I2C_Arch->PortSDA.GPIOx, ec_InvalidArgument); // ERROR: configured GPIO cannot be used in current remapping layout!
              Assert(Port.Pin   == I2C_Arch->PortSDA.Pin,   ec_InvalidArgument); // ERROR: configured Pin  cannot be used in current remapping layout!
  #else
              I2C_Arch->PortSDA.GPIOx = NULL;
  #endif
  #ifdef TTC_I2C2_SCL
              stm32_gpio_variable(&Port, TTC_I2C2_SCL);
              Assert(Port.GPIOx == I2C_Arch->PortSCL.GPIOx, ec_InvalidArgument); // ERROR: configured GPIO cannot be used in current remapping layout!
              Assert(Port.Pin   == I2C_Arch->PortSCL.Pin,   ec_InvalidArgument); // ERROR: configured Pin  cannot be used in current remapping layout!
  #else
              I2C_Arch->PortSCL.GPIOx = NULL;
  #endif
  #ifdef TTC_I2C2_SMBAL
              if (I2C_Generic->Flags.Bits.SMBus_Alert) {
                  stm32_gpio_variable(&Port, TTC_I2C2_SMBAL);
                  Assert(Port.GPIOx == I2C_Arch->PortSMBAL.GPIOx, ec_InvalidArgument); // ERROR: configured GPIO cannot be used in current remapping layout!
                  Assert(Port.Pin   == I2C_Arch->PortSMBAL.Pin,   ec_InvalidArgument); // ERROR: configured Pin  cannot be used in current remapping layout!
              }
  #else
              I2C_Arch->PortSMBAL.GPIOx = NULL;
  #endif
              break;
          }
        default: Assert(0, ec_UNKNOWN); break; // No TTC_I2Cn defined! Check your makefile.100_board_* file!
            
        }
    }

    I2C_Arch->I2C_Generic = I2C_Generic;
    ttc_i2c_errorcode_e Error = _stm32_i2c_reinit_i2c(I2C_Arch);
    if (Error != tie_OK)  return Error;

    if (0) { // ToDo: setup interrupts

    }

    stm32_I2C_Configs[I2C_Index-1] = I2C_Arch;

    return tie_OK;
}
ttc_i2c_errorcode_e stm32_i2c_write_register( u8_t I2C_Index, const u16_t TargetAddress, const u8_t Register, ttc_i2c_register_type_e RegisterType, const u8_t Byte) {
    return stm32_i2c_write_registers(I2C_Index, TargetAddress, Register, RegisterType, &Byte, 1);
}
ttc_i2c_errorcode_e stm32_i2c_write_registers(u8_t I2C_Index, const u16_t TargetAddress, const u8_t Register, ttc_i2c_register_type_e RegisterType, const u8_t* Buffer, u16_t Amount) {
  Assert(I2C_Index > 0, ec_NULL);
  stm32_i2c_architecture_t* I2C_Arch = stm32_I2C_Configs[I2C_Index-1];
  Assert(I2C_Arch != NULL, ec_NULL);
  I2C_t* I2C_Base = I2C_Arch->Base;
  Assert(I2C_Base != NULL, ec_NULL);

  if (tie_OK != _stm32_i2c_wait_for_flag(I2C_Arch, I2C_FLAG_BUSY, RESET) )                      return _stm32_i2c_close_bus(I2C_Arch, tie_TimeOut_Flag_Busy);
  if (tie_OK != _stm32_i2c_condition_start(I2C_Arch) )                                          return _stm32_i2c_close_bus(I2C_Arch, tie_TimeOut_StartCondition);
  if (tie_OK != _stm32_i2c_send_target_address(I2C_Arch, TargetAddress, tid_WritingToSlave) )   return _stm32_i2c_close_bus(I2C_Arch, tie_TimeOut_Master_ModeSelect);
  if (tie_OK != _stm32_i2c_send_register_address(I2C_Arch, Register, RegisterType) )            return _stm32_i2c_close_bus(I2C_Arch, tie_TimeOut_Master_RegisterSend); //D
  if (tie_OK != _stm32_i2c_send_bytes(I2C_Arch, Buffer, Amount) )                               return _stm32_i2c_close_bus(I2C_Arch, tie_TimeOut_Master_BufferSend);

  return _stm32_i2c_close_bus(I2C_Arch, tie_OK);
}
ttc_i2c_errorcode_e stm32_i2c_read_register(  u8_t I2C_Index, const u16_t TargetAddress, const u8_t Register, ttc_i2c_register_type_e RegisterType, u8_t* Byte) {
    RegisterType = RegisterType; // avoids warning: unused parameter 'RegisterType'

    Assert(I2C_Index > 0, ec_NULL);
    stm32_i2c_architecture_t* I2C_Arch = stm32_I2C_Configs[I2C_Index-1];
    Assert(I2C_Arch != NULL, ec_NULL);
    I2C_t* I2C_Base = I2C_Arch->Base;
    Assert(I2C_Base != NULL, ec_NULL);

    u32_t MyTimeOut = I2C_Arch->I2C_Generic->TimeOut;
    I2C_TypeDef *I2C_Module = (I2C_TypeDef *) I2C_Arch->Base;
    I2C_AcknowledgeConfig(I2C_Module, ENABLE);

    /* While the bus is busy */
    MyTimeOut = I2C_Arch->I2C_Generic->TimeOut;
    while (I2C_GetFlagStatus(I2C_Module, I2C_FLAG_BUSY)) {

        if (MyTimeOut-- == 0) {
            return tie_TimeOut_Flag_Busy;
        }
        ttc_task_yield();
    }

    /* Send START condition */
    I2C_GenerateSTART(I2C_Module, ENABLE);

    /* Test on EV5 and clear it */
    MyTimeOut = I2C_Arch->I2C_Generic->TimeOut;
    while (!I2C_CheckEvent(I2C_Module, I2C_EVENT_MASTER_MODE_SELECT)) {
        if (MyTimeOut-- == 0) {
            return tie_TimeOut_Master_ModeSelect;
        }
        ttc_task_yield();
    }

    /* Send address for write */
    I2C_Send7bitAddress(I2C_Module, TargetAddress<<1, I2C_Direction_Transmitter);

    /* Test on EV6 and clear it */
    MyTimeOut = I2C_Arch->I2C_Generic->TimeOut;
    while (!I2C_CheckEvent(I2C_Module,
                           I2C_EVENT_MASTER_TRANSMITTER_MODE_SELECTED)) {
        if (MyTimeOut-- == 0) {
            return tie_TimeOut_Master_TransmitterMode;
        }
        ttc_task_yield();
    }

    //uSleep(100);
    /* Clear EV6 by setting again the PE bit */
    I2C_Cmd(I2C_Module, ENABLE);

    /* Send the internal address to write to */
    I2C_SendData(I2C_Module, Register);

    //Send STOP Condition
    //I2C_GenerateSTOP(I2C_Module, ENABLE);

    /* While the bus is busy */
    //while (I2C_GetFlagStatus(I2C_Module, I2C_FLAG_BUSY))
    //	;

    /* Send START condition */
    I2C_GenerateSTART(I2C_Module, ENABLE);

    /* Test on EV5 and clear it */
    MyTimeOut = I2C_Arch->I2C_Generic->TimeOut;
    while (!I2C_CheckEvent(I2C_Module, I2C_EVENT_MASTER_MODE_SELECT)) {
        if (MyTimeOut-- == 0) {
            return tie_TimeOut_Master_ModeSelect;
        }
        ttc_task_yield();
    }

    /* Send address for write */
    I2C_Send7bitAddress(I2C_Module, TargetAddress<<1, I2C_Direction_Receiver);
    /* Test on EV6 and clear it */
    MyTimeOut = I2C_Arch->I2C_Generic->TimeOut;
    while (!I2C_CheckEvent(I2C_Module, I2C_EVENT_MASTER_RECEIVER_MODE_SELECTED)) {
        if (MyTimeOut-- == 0) {
            return tie_TimeOut_Master_TransmitterMode;
        };
        ttc_task_yield();
    }
    //Disable Acknowledgement
    I2C_AcknowledgeConfig(I2C_Module, DISABLE);

    //I2C_FLAG_ADDR
    I2C_GetFlagStatus(I2C_Module, I2C_FLAG_ADDR);

    //Send STOP Condition
    I2C_GenerateSTOP(I2C_Module, ENABLE);

    /* Clear EV6 by setting again the PE bit */
    //I2C_Cmd(I2C_Module, ENABLE);


    // Test on EV7 and clear it
    MyTimeOut = I2C_Arch->I2C_Generic->TimeOut;
    while (!I2C_CheckEvent(I2C_Module, I2C_EVENT_MASTER_BYTE_RECEIVED)) {
        if (MyTimeOut-- == 0) {
            return tie_TimeOut_Master_ByteRead;
        };
        ttc_task_yield();
    }

    // uSleep(150);
    /* Read a byte from the I2C */
    *Byte = I2C_ReceiveData(I2C_Module);

    /* Enable Acknowledgement to be ready for another reception */
    I2C_AcknowledgeConfig(I2C_Module, ENABLE);

    return tie_OK;
}
ttc_i2c_errorcode_e stm32_i2c_read_registers( u8_t I2C_Index, const u16_t TargetAddress, const u8_t Register, ttc_i2c_register_type_e RegisterType, u8_t* Buffer, u16_t Amount, u16_t* AmountRead) {
    Assert(I2C_Index > 0, ec_NULL);
    stm32_i2c_architecture_t* I2C_Arch = stm32_I2C_Configs[I2C_Index-1];
    Assert(I2C_Arch != NULL, ec_NULL);
    I2C_t* I2C_Base = I2C_Arch->Base;
    Assert(I2C_Base != NULL, ec_NULL);

    if (0) { // fast implementation: issue one multi-byte read (DOES NOT WORK)

        if (tie_OK != _stm32_i2c_wait_for_flag(I2C_Arch, I2C_FLAG_BUSY, RESET) )                      return _stm32_i2c_close_bus(I2C_Arch, tie_TimeOut_Flag_Busy);
        if (tie_OK != _stm32_i2c_condition_start(I2C_Arch) )                                          return _stm32_i2c_close_bus(I2C_Arch, tie_TimeOut_StartCondition);
        if (tie_OK != _stm32_i2c_send_target_address(I2C_Arch, TargetAddress, tid_WritingToSlave) )   return _stm32_i2c_close_bus(I2C_Arch, tie_TimeOut_Master_ModeSelect);
        if (tie_OK != _stm32_i2c_send_register_address(I2C_Arch, Register, RegisterType) )            return _stm32_i2c_close_bus(I2C_Arch, tie_TimeOut_Master_RegisterSend); //D

        if (1) {
            if (tie_OK != _stm32_i2c_condition_start(I2C_Arch) )                                          return _stm32_i2c_close_bus(I2C_Arch, tie_TimeOut_StartCondition);
            if (tie_OK != _stm32_i2c_send_target_address(I2C_Arch, TargetAddress, tid_ReadingFromSlave) ) return _stm32_i2c_close_bus(I2C_Arch, tie_TimeOut_Master_ModeSelect);
            if (tie_OK != _stm32_i2c_condition_start(I2C_Arch) )                                          return _stm32_i2c_close_bus(I2C_Arch, tie_TimeOut_StartCondition);
            if (tie_OK != _stm32_i2c_read_bytes(I2C_Arch, Buffer, Amount, AmountRead) )                   return _stm32_i2c_close_bus(I2C_Arch, tie_TimeOut_Master_ByteRead);
        }
        else {
            if (tie_OK != _stm32_i2c_send_bytes(I2C_Arch, Buffer, Amount) )                               return _stm32_i2c_close_bus(I2C_Arch, tie_TimeOut_Master_BufferSend);
        }

        return _stm32_i2c_close_bus(I2C_Arch, tie_OK);

    }
    else {   // slow implementation: issue multiple singly-byte reads
        ttc_i2c_errorcode_e ReturnCode = tie_OK;
        while (Amount > 0) {
            ReturnCode = stm32_i2c_read_register(I2C_Index, TargetAddress, Register, RegisterType, Buffer);
            if (ReturnCode == tie_OK) {
                Buffer++;
                AmountRead++;
            }
            else return ReturnCode;
        }
    }

    return tie_OK;
}

//} Function definitions
//{ private functions (ideally) *************************************************

               u8_t _stm32_i2c_get_pins(u8_t I2C_Index, stm32_i2c_architecture_t* I2C_Arch) {
    u8_t Layout = 0;
    stm32_Port_t PortSDA;

    switch (I2C_Index) { // check if board has defined the SDA-pin of I2C to init
#ifdef TTC_I2C1_SDA
    case 1:
        stm32_gpio_variable(&PortSDA, TTC_I2C1_SDA);

        break;
#endif
#ifdef TTC_I2C2_SDA
    case 2:
        stm32_gpio_variable(&PortSDA, TTC_I2C2_SDA);
        break;
#endif
    default:
        Assert(0, ec_UNKNOWN);   // No TTC_I2Cn_SDA defined! Check your makefile.100_board_* file!
        break;
    }

    switch ( (u32_t) I2C_Arch->Base ) { // determine pin layout (-> RM0008 p.176)
    case (u32_t) I2C1: { // I2C1
        if (PortSDA.Pin == 6) Layout = 1;
        else                  Layout = 0;

        if (Layout) { // alternate pin remapping (-> ST RM0041 STM32F100xx advanced ARM-based 32-bit MCUs p.118)
            I2C_Arch->PortSDA.GPIOx   = GPIOB;
            I2C_Arch->PortSDA.Pin     = 9;
            I2C_Arch->PortSCL.GPIOx   = GPIOB;
            I2C_Arch->PortSCL.Pin     = 8;
        }
        else {        // no pin remapping
            I2C_Arch->PortSDA.GPIOx   = GPIOB;
            I2C_Arch->PortSDA.Pin     = 7;
            I2C_Arch->PortSCL.GPIOx   = GPIOB;
            I2C_Arch->PortSCL.Pin     = 6;
        }
        I2C_Arch->PortSMBAL.GPIOx = GPIOB;
        I2C_Arch->PortSMBAL.Pin   = 5;

        break;
    }
    case (u32_t) I2C2: { // I2C2

        I2C_Arch->PortSDA.GPIOx   = GPIOB;
        I2C_Arch->PortSDA.Pin     = 11;
        I2C_Arch->PortSCL.GPIOx   = GPIOB;
        I2C_Arch->PortSCL.Pin     = 10;
        I2C_Arch->PortSMBAL.GPIOx = GPIOB;
        I2C_Arch->PortSMBAL.Pin   = 12;

        break;
    }
    default: { return tie_DeviceNotFound; }
    }

    return Layout;
}
ttc_i2c_errorcode_e _stm32_i2c_wait_for_flag(stm32_i2c_architecture_t* I2C_Arch, u32_t Flag, FlagStatus WaitForState) {
    Assert(I2C_Arch != NULL, ec_NULL);
    I2C_t* I2C_Base = I2C_Arch->Base;
    Assert(I2C_Base != NULL, ec_NULL);
    u16_t MyTimeOut = I2C_Arch->I2C_Generic->TimeOut;

    if (MyTimeOut) {
        while( WaitForState != I2C_GetFlagStatus( (I2C_TypeDef*) I2C_Base, Flag) && (MyTimeOut > 0) ) {
            MyTimeOut--;
            ttc_task_yield();
        }
        if (MyTimeOut == 0)
            return tie_TimeOut;
    }
    else
        while( WaitForState != I2C_GetFlagStatus( (I2C_TypeDef*) I2C_Base, Flag) ) ttc_task_yield();

    return tie_OK;
}
ttc_i2c_errorcode_e _stm32_i2c_wait_for_event(stm32_i2c_architecture_t* I2C_Arch, u32_t Event) {
    Assert(I2C_Arch != NULL, ec_NULL);
    I2C_t* I2C_Base = I2C_Arch->Base;
    Assert(I2C_Base != NULL, ec_NULL);
    u16_t MyTimeOut = I2C_Arch->I2C_Generic->TimeOut;

    if (MyTimeOut) {
        if (1) {
            while( (SUCCESS != I2C_CheckEvent( (I2C_TypeDef*) I2C_Base, Event)) && (MyTimeOut > 0) ) {
                MyTimeOut--;
                ttc_task_yield();
            }
            if (MyTimeOut == 0)  return tie_TimeOut;
        }
        else {
            ttc_task_begin_criticalsection(_stm32_i2c_wait_for_event);
            while (SUCCESS != I2C_CheckEvent( (I2C_TypeDef*) I2C_Base, Event));
            ttc_task_end_criticalsection();
        }
    }
    else
        while( SUCCESS != I2C_CheckEvent( (I2C_TypeDef*) I2C_Base, Event) )
            ttc_task_yield();

    return tie_OK;
}
ttc_i2c_errorcode_e _stm32_i2c_read_bytes(stm32_i2c_architecture_t* I2C_Arch, u8_t* Buffer, u16_t Amount, u16_t* AmountRead) {
    Assert(I2C_Arch != NULL, ec_NULL);
    I2C_t* I2C_Base = I2C_Arch->Base;
    Assert(I2C_Base != NULL, ec_NULL);

    ttc_i2c_errorcode_e Error;

    _stm32_i2c_condition_stop(I2C_Arch);
    if (0) ttc_task_begin_criticalsection(_stm32_i2c_read_bytes);
    while (Amount-- > 0) {
        Error = _stm32_i2c_read_byte(I2C_Arch, Buffer);
        if (Error) return Error;
        (*AmountRead)++;
        Buffer++;
    }
    if (0) ttc_task_end_criticalsection();

    return tie_OK;
}
ttc_i2c_errorcode_e _stm32_i2c_read_byte(stm32_i2c_architecture_t* I2C_Arch, u8_t* Buffer) {
    Assert(I2C_Arch != NULL, ec_NULL);
    I2C_t* I2C_Base = I2C_Arch->Base;
    Assert(I2C_Base != NULL, ec_NULL);

    // clear all flags
    (void) I2C_Base->SR1;
    (void) I2C_Base->SR2;

    u16_t MyTimeOut = I2C_Arch->I2C_Generic->TimeOut;
    while ( (MyTimeOut > 0) && (I2C_Base->SR1.RXNE) ) {
        ttc_task_yield();
        MyTimeOut--;
    }
    if (MyTimeOut > 0) {
        *Buffer = *( (u8_t*) &(I2C_Arch->Base->DR) );
        return tie_OK;
    }
    return tie_TimeOut;
}
ttc_i2c_errorcode_e _stm32_i2c_send_register_address(stm32_i2c_architecture_t* I2C_Arch, u32_t Register, ttc_i2c_register_type_e RegisterType) {
    Assert(I2C_Arch != NULL, ec_NULL);
    I2C_t* I2C_Base = I2C_Arch->Base;
    Assert(I2C_Base != NULL, ec_NULL);

    u8_t AmountBytes = 0;
    u8_t Buffer[4];

    switch (RegisterType) { // send register adress (MSB first)

    case tiras_MSB_First_32Bit:
        Buffer[0] = (u8_t) (Register & 0xff000000) >> 24;
        Buffer[1] = (u8_t) (Register & 0x00ff0000) >> 16;
        Buffer[2] = (u8_t) (Register & 0x0000ff00) >> 8;
        Buffer[3] = (u8_t) Register & 0x000000ff;
        AmountBytes = 4;
        break;
    case tiras_MSB_First_24Bit:
        Buffer[0] = (u8_t) (Register & 0x00ff0000) >> 16;
        Buffer[1] = (u8_t) (Register & 0x0000ff00) >> 8;
        Buffer[2] = (u8_t) Register & 0x000000ff;
        AmountBytes = 3;
        break;

    case tiras_MSB_First_16Bit:
        Buffer[0] = (u8_t) (Register & 0x0000ff00) >> 8;
        Buffer[1] = (u8_t) Register & 0x000000ff;
        AmountBytes = 2;
        break;

    case tiras_MSB_First_8Bit:
        Buffer[0] = (u8_t) Register & 0x000000ff;
        AmountBytes = 1;
        break;

    case tiras_LSB_First_32Bit:
        Buffer[0] = (u8_t) Register & 0x000000ff;
        Buffer[1] = (u8_t) (Register & 0x0000ff00) >> 8;
        Buffer[2] = (u8_t) (Register & 0x00ff0000) >> 16;
        Buffer[3] = (u8_t) (Register & 0xff000000) >> 24;
        AmountBytes = 4;
        break;
    case tiras_LSB_First_24Bit:
        Buffer[0] = (u8_t) Register & 0x000000ff;
        Buffer[1] = (u8_t) (Register & 0x0000ff00) >> 8;
        Buffer[2] = (u8_t) (Register & 0x00ff0000) >> 16;
        AmountBytes = 3;
        break;

    case tiras_LSB_First_16Bit:
        Buffer[0] = (u8_t) Register & 0x000000ff;
        Buffer[1] = (u8_t) (Register & 0x0000ff00) >> 8;
        AmountBytes = 2;
        break;

    case tiras_LSB_First_8Bit:
        Buffer[0] = (u8_t) Register & 0x000000ff;
        AmountBytes = 1;
        break;

    default: Assert(0, ec_InvalidArgument); // unknown register size
    }

    return _stm32_i2c_send_bytes(I2C_Arch, Buffer, AmountBytes);
}
ttc_i2c_errorcode_e _stm32_i2c_send_bytes(stm32_i2c_architecture_t* I2C_Arch, const u8_t* Buffer, u16_t Amount) {
    Assert(I2C_Arch != NULL, ec_NULL);
    I2C_t* I2C_Base = I2C_Arch->Base;
    Assert(I2C_Base != NULL, ec_NULL);

    ttc_i2c_errorcode_e Error;

    while (Amount-- > 0) {
        Error = _stm32_i2c_send_byte(I2C_Arch, *Buffer++);
        if (Error)
            return Error;
    }

    return tie_OK;
}
ttc_i2c_errorcode_e _stm32_i2c_send_byte(stm32_i2c_architecture_t* I2C_Arch, const u8_t Byte) {
    Assert(I2C_Arch != NULL, ec_NULL);
    I2C_t* I2C_Base = I2C_Arch->Base;
    Assert(I2C_Base != NULL, ec_NULL);

    if (USE_OWN_I2C_CODE) {
        *( (u8_t*) &(I2C_Base->DR) ) = Byte;
        I2C_SR1_t SR1;
        I2C_SR2_t SR2;

        //X I2C_Base->CR1.PE = 1;  // reset EV6
        if (0) {
            u16_t MyTimeOut = I2C_Arch->I2C_Generic->TimeOut;
            while ( (MyTimeOut > 0) && (! SR1.TxE) ) {
                MyTimeOut--;

                // Read from SR1 and SR2 register to clear all flags
                SR1 = I2C_Base->SR1;
                SR2 = I2C_Base->SR2;

                ttc_task_yield();
            }
            if (MyTimeOut == 0)
                return tie_TimeOut_Master_TransmitterMode;
        }
        else {
            if (tie_OK != _stm32_i2c_wait_for_event(I2C_Arch, I2C_EVENT_MASTER_BYTE_TRANSMITTED) )
                return tie_TimeOut_Master_ByteTransmitted;
        }
    }
    else {
        I2C_SendData( (I2C_TypeDef*) I2C_Arch->Base, Byte);
        if (tie_OK != _stm32_i2c_wait_for_event(I2C_Arch, I2C_EVENT_MASTER_BYTE_TRANSMITTED) )
            return tie_TimeOut_Master_ByteTransmitted;
    }

    return tie_OK;
}
ttc_i2c_errorcode_e _stm32_i2c_close_bus(stm32_i2c_architecture_t* I2C_Arch, ttc_i2c_errorcode_e Status) {
    Assert(I2C_Arch != NULL, ec_NULL);
    I2C_t* I2C_Base = I2C_Arch->Base;
    Assert(I2C_Base != NULL, ec_NULL);

    if (0) {
        I2C_SR1_t SR1 = I2C_Base->SR1;
        if (SR1.TIMEOUT)  Status = tie_TimeOut;
        if (SR1.BERR)     Status = tie_BusError;
        if (SR1.ARLO)     Status = tie_Master_ArbitrationLost;
        if (SR1.AF)       Status = tie_AcknowledgeFailure;
        if (SR1.PECERR)   Status = tie_PEC_Error;
        if (SR1.SMBALERT) Status = tie_SMBusAlert;
    }

    if (Status == tie_OK) {
        I2C_GenerateSTOP( (I2C_TypeDef*) I2C_Base, ENABLE);
    }
    else { // error-status: send out zero bits to reset all slaves
        _stm32_i2c_reinit_i2c(I2C_Arch);
        if (0) { // try to reset slaves
            // _stm32_i2c_condition_start(I2C_Arch);

            // send out zero byte to flush slaves shift register
            u8_t Zero = 0;
            _stm32_i2c_send_bytes(I2C_Arch, &Zero, 1);

            // _stm32_i2c_condition_stop(I2C_Arch);
        }
        I2C_Arch->I2C_Generic->ErrorCount++;
    }

    // Read from SR1 and SR2 register to clear all flags
    (void) I2C_Base->SR1;
    (void) I2C_Base->SR2;

    return Status;
}
ttc_i2c_errorcode_e _stm32_i2c_reinit_i2c(stm32_i2c_architecture_t* I2C_Arch) {
    Assert(I2C_Arch != NULL, ec_NULL);
    ttc_i2c_generic_t* I2C_Generic = I2C_Arch->I2C_Generic;
    Assert(I2C_Generic != NULL, ec_NULL);
    I2C_t* I2C_Base = I2C_Arch->Base;
    Assert(I2C_Base != NULL, ec_NULL);

    if (1) { // deconfigure GPIOs

        if (I2C_Arch->PortSDA.GPIOx)
            stm32_gpio_init_variable(&(I2C_Arch->PortSDA),    tgm_input_floating, GPIO_Speed_10MHz);

        if (I2C_Arch->PortSCL.GPIOx)
            stm32_gpio_init_variable(&(I2C_Arch->PortSCL),    tgm_input_floating, GPIO_Speed_10MHz);

        if (I2C_Arch->PortSMBAL.GPIOx && I2C_Generic->Flags.Bits.SMBus_Alert)
            stm32_gpio_init_variable(&(I2C_Arch->PortSMBAL),  tgm_input_floating,  GPIO_Speed_10MHz);
    }
    I2C_Base->CR1.SWRST = 1;
    I2C_DeInit( (I2C_TypeDef*) I2C_Base);
    if (1) { // activate clocks
        switch ( (u32_t) I2C_Base ) {
          case (u32_t) I2C1: {
              if (I2C_Generic->Layout) { // GPIOB + GPIOA
                  RCC_APB2PeriphClockCmd(RCC_APB1Periph_I2C1 | RCC_APB2Periph_GPIOA | RCC_APB2Periph_GPIOB | RCC_APB2Periph_AFIO, ENABLE);
              }
              else {                     // GPIOA
                  RCC_APB2PeriphClockCmd(RCC_APB1Periph_I2C1 | RCC_APB2Periph_GPIOA | RCC_APB2Periph_AFIO, ENABLE);
              }
              break;
          }
          case (u32_t) I2C2: {
              RCC_APB1PeriphClockCmd(RCC_APB1Periph_I2C2, ENABLE);
              RCC_APB2PeriphClockCmd(RCC_APB2Periph_GPIOB | RCC_APB2Periph_AFIO, ENABLE);
              break;
          }
          default: Assert(0, ec_UNKNOWN); break; // should never occur!
        }
    }
    if (1) { // reset I2C device
        switch ( (u32_t) I2C_Base ) {
        case (u32_t) I2C1: {
        default:
            RCC_APB1PeriphResetCmd(RCC_APB1Periph_I2C1, ENABLE);  // Reset I2C
            RCC_APB1PeriphResetCmd(RCC_APB1Periph_I2C1, DISABLE); // Release I2C from reset state
            break;
        }
        case (u32_t) I2C2: {
            RCC_APB1PeriphResetCmd(RCC_APB1Periph_I2C2, ENABLE);  // Reset I2C
            RCC_APB1PeriphResetCmd(RCC_APB1Periph_I2C2, DISABLE); // Release I2C from reset state
            break;
        }
        }

    }
    if (1) { // apply remap layout

        switch ( (u32_t) I2C_Base ) {
          case (u32_t) I2C1: { // I2C1
              if (I2C_Generic->Layout)
                  GPIO_PinRemapConfig(GPIO_Remap_I2C1, ENABLE);
              else
                  GPIO_PinRemapConfig(GPIO_Remap_I2C1, DISABLE);

              break;
          }
          default: { break; }
        }
    }

    I2C_Cmd( (I2C_TypeDef*) I2C_Base, ENABLE);

    if (1) { // init I2C after enabling it

        I2C_InitTypeDef I2C_MyInit;
        memset( &I2C_MyInit, 0, sizeof(I2C_InitTypeDef) );

        I2C_MyInit.I2C_OwnAddress1 = I2C_Generic->OwnAddress;
        I2C_MyInit.I2C_ClockSpeed  = I2C_Generic->ClockSpeed;
        if (I2C_Generic->Flags.Bits.SMBus) {
            if (I2C_Generic->Flags.Bits.SMBus_Host)
                I2C_MyInit.I2C_Mode = I2C_Mode_SMBusHost;
            else
                I2C_MyInit.I2C_Mode = I2C_Mode_SMBusDevice;
        }
        else  I2C_MyInit.I2C_Mode = I2C_Mode_I2C;

        if (I2C_Generic->Flags.Bits.Acknowledgement)  I2C_MyInit.I2C_Ack = I2C_Ack_Enable;
        else                                          I2C_MyInit.I2C_Ack = I2C_Ack_Disable;

        if (I2C_Generic->Flags.Bits.Adress_10Bit)     I2C_MyInit.I2C_AcknowledgedAddress = I2C_AcknowledgedAddress_10bit;
        else                                          I2C_MyInit.I2C_AcknowledgedAddress = I2C_AcknowledgedAddress_7bit;

        if (I2C_Generic->Flags.Bits.DutyCycle_16_9)   I2C_MyInit.I2C_DutyCycle = I2C_DutyCycle_16_9;
        else                                          I2C_MyInit.I2C_DutyCycle = I2C_DutyCycle_2;

        I2C_Init((I2C_TypeDef*) I2C_Base, &I2C_MyInit);

        // do stuff that was not done by standard_peripherals_library
        I2C_Base->CCR.Fast = I2C_Generic->Flags.Bits.ModeFast;
    }
    if (1) { // configure GPIOs

        if (0) { // test pins
            stm32_gpio_init_variable(&(I2C_Arch->PortSDA), tgm_output_open_drain, GPIO_Speed_10MHz);
            stm32_gpio_init_variable(&(I2C_Arch->PortSCL), tgm_output_open_drain, GPIO_Speed_10MHz);

            while (1) {

                stm32_gpio_set_variable(&(I2C_Arch->PortSDA));
                stm32_gpio_set_variable(&(I2C_Arch->PortSCL));
                uSleep(10);
                stm32_gpio_clr_variable(&(I2C_Arch->PortSCL));
                uSleep(10);
                stm32_gpio_clr_variable(&(I2C_Arch->PortSDA));
                stm32_gpio_set_variable(&(I2C_Arch->PortSCL));
                uSleep(10);
                stm32_gpio_clr_variable(&(I2C_Arch->PortSCL));
                uSleep(10);
            }
        }

        if (I2C_Arch->PortSDA.GPIOx)
            stm32_gpio_init_variable(&(I2C_Arch->PortSDA), tgm_alternate_function_open_drain, GPIO_Speed_10MHz);

        if (I2C_Arch->PortSCL.GPIOx)
            stm32_gpio_init_variable(&(I2C_Arch->PortSCL), tgm_alternate_function_open_drain, GPIO_Speed_10MHz);

        if (I2C_Arch->PortSMBAL.GPIOx && I2C_Generic->Flags.Bits.SMBus_Alert)
            stm32_gpio_init_variable(&(I2C_Arch->PortSMBAL),  tgm_alternate_function_open_drain,  GPIO_Speed_10MHz);
    }

    return tie_OK;
}
ttc_i2c_errorcode_e _stm32_i2c_condition_start(stm32_i2c_architecture_t* I2C_Arch) {
    Assert(I2C_Arch != NULL, ec_NULL);
    I2C_t* I2C_Base = I2C_Arch->Base;
    Assert(I2C_Base != NULL, ec_NULL);

    I2C_Base->CR1.START = 1;
    u16_t MyTimeOut = I2C_Arch->I2C_Generic->TimeOut;
    while ( (MyTimeOut > 0) && I2C_Base->CR1.START) {
        MyTimeOut--;
        ttc_task_yield();
    }

    if (MyTimeOut > 0)  return tie_OK;

    return tie_TimeOut_StopCondition;
}
ttc_i2c_errorcode_e _stm32_i2c_condition_stop(stm32_i2c_architecture_t* I2C_Arch) {
    Assert(I2C_Arch != NULL, ec_NULL);
    I2C_t* I2C_Base = I2C_Arch->Base;
    Assert(I2C_Base != NULL, ec_NULL);

    I2C_Base->CR1.STOP = 1;
    u16_t MyTimeOut = I2C_Arch->I2C_Generic->TimeOut;
    while ( (MyTimeOut > 0) && I2C_Base->CR1.STOP) {
        MyTimeOut--;
        ttc_task_yield();
    }

    if (MyTimeOut > 0)  return tie_OK;

    return tie_TimeOut_StopCondition;
}
ttc_i2c_errorcode_e _stm32_i2c_send_target_address(stm32_i2c_architecture_t* I2C_Arch, u16_t TargetAddress, ttc_i2c_direction_e Direction) {
    Assert(I2C_Arch != NULL, ec_NULL);
    I2C_t* I2C_Base = I2C_Arch->Base;
    Assert(I2C_Base != NULL, ec_NULL);

    ttc_i2c_Address_7Bit_u Data;
    if (Direction == tid_WritingToSlave)
        Data.Fields.ReadMode = 0;               // writing to slave
    else
        Data.Fields.ReadMode = 1;               // reading from slave

    Data.Fields.Address = TargetAddress;

    if (0) {
        if (0) I2C_SendData( (I2C_TypeDef*) I2C_Base, Data.Byte);
        else   *( (u8_t*) &(I2C_Base->DR) ) = Data.Byte;
        u16_t MyTimeOut = I2C_Arch->I2C_Generic->TimeOut;
        while ( (MyTimeOut > 0) && (!I2C_Base->SR1.ADDR) ) { // EV6
            MyTimeOut--;
            ttc_task_yield();
        }
        if (MyTimeOut == 0) return tie_TimeOut_Master_TransmitterMode;
        I2C_Base->CR1.PE = 1;  // reset EV6
        MyTimeOut = I2C_Arch->I2C_Generic->TimeOut;
        while ( (MyTimeOut > 0) && (!I2C_Base->SR1.TxE) ) { // EV8_1
            MyTimeOut--;
            // Read from SR1 and SR2 register to clear all flags
            (void) I2C_Base->SR1;
            (void) I2C_Base->SR2;
            ttc_task_yield();
        }

        if (MyTimeOut == 0) return tie_TimeOut_Master_TransmitterMode;

        if (tie_OK != _stm32_i2c_wait_for_event(I2C_Arch, I2C_EVENT_MASTER_TRANSMITTER_MODE_SELECTED) )
            return tie_TimeOut_Master_TransmitterMode;
    }
    else {
       *( (u8_t*) &(I2C_Base->DR) ) = Data.Byte;
        u16_t MyTimeOut = I2C_Arch->I2C_Generic->TimeOut;
        while ( (MyTimeOut-- > 0) && (!I2C_CheckEvent( (I2C_TypeDef*) I2C_Base, I2C_EVENT_MASTER_TRANSMITTER_MODE_SELECTED)) )
            ttc_task_yield();
        if (MyTimeOut == 0)
            return tie_TimeOut_Master_TransmitterMode;
    }

    return tie_OK;
}

//} private functions
