/*{ stm32_interrupt::.c -------------------------------------------------------------

                      The ToolChain
                      
   Interrupt configuration
   for ARM CortexM3 microcontrollers.
   Currently tested to comply to STM32F10x types.
   
   written by Gregor Rebel 2012
   
   Note: See ttc_interrupt.h for description of architecture independent INTERRUPT implementation.

}*/

#include "stm32_interrupt.h"

// stores pointers to structs of interrupt service routines for each functional unit
ttc_array_dynamic(stm32_usart_isrs_t*, stm32_usart_isrs);
ttc_array_dynamic(stm32_gpio_isrs_t*,  stm32_gpio_isrs);


#ifdef EXTENSION_400_stm32_usb_fs_host_lib
    extern USB_OTG_CORE_HANDLE          USB_OTG_Core_dev;
    extern USBH_HOST                    USB_Host;
#endif
//{ Function definitions --------------------------------------------------------

ttc_interrupt_errorcode_e stm32_interrupt_init(ttc_interrupt_type_e Type, physical_index_t PhysicalIndex, void (*ISR)(physical_index_t, void*), void* Argument, void (**OldISR)(physical_index_t, void*), void** OldArgument) {

    // Important: enums must checked in order from high to low!

    // Highest enum
    if (Type > tit_First_USART)
        return stm32_interrupt_init_usart(Type, PhysicalIndex, ISR, Argument, OldISR, OldArgument);
    if (Type > tit_First_SPI)
        return tine_NotImplemented;
    if (Type > tit_First_RADIO)
        return tine_NotImplemented;
    if (Type > tit_First_TIMER)
        return tine_NotImplemented;
    if (Type > tit_First_I2C)
        return tine_NotImplemented;
    if (Type > tit_First_GPIO)
        return stm32_interrupt_init_gpio(Type, PhysicalIndex, ISR, Argument, OldISR, OldArgument);
    // Lowest enum

    Assert_Interrupt(0, ec_InvalidArgument); // unknown Type given!
    return tine_NotImplemented;    // just to make compiler happy
}
ttc_interrupt_errorcode_e stm32_interrupt_init_usart(ttc_interrupt_type_e Type, physical_index_t PhysicalIndex, void (*ISR)(physical_index_t, void*), void* Argument, void (**OldISR)(physical_index_t, void*), void** OldArgument) {
    if ( ttc_array_is_null(stm32_usart_isrs) ) { // first init of an USART-irq: create list of pointers
        ttc_array_allocate(stm32_usart_isrs, TTC_USART_MAX_AMOUNT);
    }
    if (A(stm32_usart_isrs, PhysicalIndex) == NULL) { // first init of irq for USART of this index: create emptx pointer table
        A(stm32_usart_isrs, PhysicalIndex) = ttc_memory_alloc_zeroed( sizeof(stm32_usart_isrs_t) );
    }

    void (**Entry_ISR)(physical_index_t, void*) = NULL;
    ttc_usart_generic_t** Entry_Argument = NULL;
    stm32_usart_isrs_t* Entry = A(stm32_usart_isrs, PhysicalIndex);
    switch (Type) { // find memory location where to store function pointer + argument

    case tit_USART_Cts:
        Entry_ISR = (void(**)(physical_index_t, void*)) &(Entry->isr_ClearToSend);
        Entry_Argument = &(Entry->Argument_ClearToSend);
        break;
    case tit_USART_Idle:
        Entry_ISR = (void(**)(physical_index_t, void*)) &(Entry->isr_IdleLine);
        Entry_Argument = &(Entry->Argument_IdleLine);
        break;
    case tit_USART_TxComplete:
        Entry_ISR = (void(**)(physical_index_t, void*)) &(Entry->isr_TransmissionComplete);
        Entry_Argument = &(Entry->Argument_TransmissionComplete);
        break;
    case tit_USART_TransmitDataEmpty:
        Entry_ISR = (void(**)(physical_index_t, void*)) &(Entry->isr_TransmitDataEmpty);
        Entry_Argument = &(Entry->Argument_TransmitDataEmpty);
        break;
    case tit_USART_RxNE:
        Entry_ISR = (void(**)(physical_index_t, void*)) &(Entry->isr_ReceiveDataNotEmpty);
        Entry_Argument = &(Entry->Argument_ReceiveDataNotEmpty);
        break;
    case tit_USART_LinBreak:
        Entry_ISR = (void(**)(physical_index_t, void*)) &(Entry->isr_LinBreakDetected);
        Entry_Argument = &(Entry->Argument_LinBreakDetected);
        break;
    case tit_USART_Error:
        Entry_ISR = (void(**)(physical_index_t, void*)) &(Entry->isr_Error);
        Entry_Argument = &(Entry->Argument_Error);
        break;
    default: break;
    }

    if (OldISR != NULL)        // save old value of function ptr
        *OldISR = *Entry_ISR;
    if (OldArgument != NULL)   // save old argument
        *OldArgument = (void*) *Entry_Argument;

    *Entry_ISR = ISR;  // establish new isr
    Assert_Interrupt(Entry_Argument, ec_NULL);
    *Entry_Argument = (ttc_usart_generic_t*) Argument;

    return tine_OK;
}
ttc_interrupt_errorcode_e stm32_interrupt_init_gpio(ttc_interrupt_type_e Type, physical_index_t PhysicalIndex, void (*ISR)(physical_index_t, void*), void* Argument, void (**OldISR)(physical_index_t, void*), void** OldArgument) {
    GPIO_TypeDef* GPIOx;
    u8_t Pin;
    stm32_gpio_from_index8(PhysicalIndex, &GPIOx, &Pin);
    Assert_Interrupt(GPIOx,    ec_InvalidArgument);
    Assert_Interrupt(Pin < TTC_GPIO_MAX_PINS, ec_InvalidArgument);

    if ( ttc_array_is_null(stm32_gpio_isrs) ) { // first init of an gpio-irq: create list of pointers
        ttc_array_allocate(stm32_gpio_isrs, 16); // CortexM3 provides 16 lines for external gpio-interrupts
    }
    if (A(stm32_gpio_isrs, Pin) == NULL) {      // first init of irq for gpio of this index: create emptx pointer table
        A(stm32_gpio_isrs, Pin) = ttc_memory_alloc_zeroed( sizeof(stm32_gpio_isrs_t) );
    }

    stm32_gpio_isrs_t* GPIO_ISR_Entry = A(stm32_gpio_isrs, Pin);

    // save previous entries
    if (OldISR)      *OldISR      = GPIO_ISR_Entry->isr;
    if (OldArgument) *OldArgument = GPIO_ISR_Entry->Argument;

    // load new entries
    GPIO_ISR_Entry->isr           = ISR;
    GPIO_ISR_Entry->Argument      = Argument;
    GPIO_ISR_Entry->PhysicalIndex = PhysicalIndex;

    EXTI_InitTypeDef EXTI_Cfg;
    NVIC_InitTypeDef NVIC_Cfg;

    u8_t GPIO_PortSource = 0;
    u8_t GPIO_PinType    = Pin;

    EXTI_StructInit(&EXTI_Cfg);

    /* Enable AFIO clock */
    RCC_APB2PeriphClockCmd(RCC_APB2Periph_AFIO, ENABLE);

    switch ( (u32_t) GPIOx ) {
#ifdef GPIOA
    case (u32_t) GPIOA:
        GPIO_PortSource = GPIO_PortSourceGPIOA;
        break;
#endif
#ifdef GPIOB
    case (u32_t) GPIOB:
        GPIO_PortSource = GPIO_PortSourceGPIOB;
        break;
#endif
#ifdef GPIOC
    case (u32_t) GPIOC:
        GPIO_PortSource = GPIO_PortSourceGPIOC;
        break;
#endif
#ifdef GPIOD
    case (u32_t) GPIOD:
        GPIO_PortSource = GPIO_PortSourceGPIOD;
        break;
#endif
#ifdef GPIOE
    case (u32_t) GPIOE:
        GPIO_PortSource = GPIO_PortSourceGPIOE;
        break;
#endif
#ifdef GPIOF
    case (u32_t) GPIOF:
        GPIO_PortSource = GPIO_PortSourceGPIOF;
        break;
#endif
#ifdef GPIOG
    case (u32_t) GPIOG:
        GPIO_PortSource = GPIO_PortSourceGPIOG;
        break;
#endif
    default:
        Assert_Interrupt(0,ec_InvalidArgument);
    }

    EXTI_Cfg.EXTI_Mode = EXTI_Mode_Interrupt;
    EXTI_Cfg.EXTI_Line = 1 << Pin;

    switch (Type) { // find memory location where to store function pointer + argument

    case tit_GPIO_Falling:
        EXTI_Cfg.EXTI_Trigger = EXTI_Trigger_Falling;
        break;

    case tit_GPIO_Rising:
        EXTI_Cfg.EXTI_Trigger = EXTI_Trigger_Rising;
        break;

    case tit_GPIO_Rising_Falling:
        EXTI_Cfg.EXTI_Trigger = EXTI_Trigger_Rising_Falling;
        break;

    default:
        Assert_Interrupt(0,ec_InvalidArgument);
        break;
    }

    //configure the External interrupt/event line mapping
    GPIO_EXTILineConfig(GPIO_PortSource, GPIO_PinType);

    EXTI_Cfg.EXTI_LineCmd = DISABLE;
    EXTI_Init(&EXTI_Cfg);

    // 32-bit architecture allows fast constant arrays with direct pointer access
    static const u8_t IRQChannel[TTC_GPIO_MAX_PINS] = { EXTI0_IRQn,
                                                        EXTI1_IRQn,
                                                        EXTI2_IRQn,
                                                        EXTI3_IRQn,
                                                        EXTI4_IRQn,
                                                        EXTI9_5_IRQn,
                                                        EXTI9_5_IRQn,
                                                        EXTI9_5_IRQn,
                                                        EXTI9_5_IRQn,
                                                        EXTI9_5_IRQn,
                                                        EXTI15_10_IRQn,
                                                        EXTI15_10_IRQn,
                                                        EXTI15_10_IRQn,
                                                        EXTI15_10_IRQn,
                                                        EXTI15_10_IRQn,
                                                        EXTI15_10_IRQn
                                                      };

    NVIC_Cfg.NVIC_IRQChannel                   = IRQChannel[Pin];
    NVIC_Cfg.NVIC_IRQChannelPreemptionPriority = 0x0F;
    NVIC_Cfg.NVIC_IRQChannelSubPriority        = 0x0F;
    NVIC_Cfg.NVIC_IRQChannelCmd                = ENABLE;
    NVIC_Init(&NVIC_Cfg);

    return tine_OK;
}
void stm32_interrupt_enable(ttc_interrupt_type_e Type, physical_index_t PhysicalIndex, BOOL EnableIRQ) {

    // Important: enums must be checked in order from high to low!

    // Highest enum
    if (Type > tit_First_USART) {
        stm32_interrupt_enable_usart(Type, PhysicalIndex, EnableIRQ);
        return;
    }
    if (Type > tit_First_SPI)
        return;
    if (Type > tit_First_RADIO)
        return;
    if (Type > tit_First_TIMER)
        return;
    if (Type > tit_First_I2C)
        return;
    if (Type > tit_First_GPIO) {
        stm32_interrupt_enable_gpio(Type, PhysicalIndex, EnableIRQ);
        return;
    }
    // Lowest enum

    Assert_USART(0, ec_InvalidConfiguration);  // Type not known
}
void stm32_interrupt_enable_usart(ttc_interrupt_type_e Type, physical_index_t PhysicalIndex, BOOL EnableIRQ) {
    NVIC_InitTypeDef NVIC_Cfg;

    /* Enable and set USARTn Interrupt to the lowest priority */
    NVIC_Cfg.NVIC_IRQChannelPreemptionPriority= 0x0F;
    NVIC_Cfg.NVIC_IRQChannelSubPriority= 0x0F;
    NVIC_Cfg.NVIC_IRQChannelCmd=ENABLE;

    USART_TypeDef* USART_Base = NULL;

    switch (PhysicalIndex) { // physical index starts at 0

    case 0: USART_Base = USART1; NVIC_Cfg.NVIC_IRQChannel = USART1_IRQn; break;
    case 1: USART_Base = USART2; NVIC_Cfg.NVIC_IRQChannel = USART2_IRQn; break;
    case 2: USART_Base = USART3; NVIC_Cfg.NVIC_IRQChannel = USART3_IRQn; break;
#ifdef UART4_IRQn
    case 3: USART_Base = UART4;  NVIC_Cfg.NVIC_IRQChannel =  UART4_IRQn; break;
#endif
#ifdef UART5_IRQn
    case 4: USART_Base = UART5;  NVIC_Cfg.NVIC_IRQChannel =  UART5_IRQn; break;
#endif
    default: { Assert_Interrupt(0, ec_InvalidConfiguration); }
    }

    USART_CR1_t CR1; *( (u16_t*) &CR1 ) = ((USART_TypeDef*) USART_Base)->CR1;
    USART_CR2_t CR2; *( (u16_t*) &CR2 ) = ((USART_TypeDef*) USART_Base)->CR2;
    USART_CR3_t CR3; *( (u16_t*) &CR3 ) = ((USART_TypeDef*) USART_Base)->CR3;

    switch (Type) { // set/ clear corresponding bits

    case tit_USART_Cts:
        if (EnableIRQ)
            Assert_Interrupt(A(stm32_usart_isrs, PhysicalIndex)->isr_ClearToSend != NULL, ec_InvalidConfiguration); // no isr configured for this interrupt!
        CR3.CTSIE  = EnableIRQ;
        break;
    case tit_USART_Idle:
        CR1.IDLEIE = EnableIRQ;
        if (EnableIRQ)
            Assert_Interrupt(A(stm32_usart_isrs, PhysicalIndex)->isr_IdleLine != NULL, ec_InvalidConfiguration); // no isr configured for this interrupt!
        break;
    case tit_USART_TxComplete:
        CR1.TCIE   = EnableIRQ;
        if (EnableIRQ)
            Assert_Interrupt(A(stm32_usart_isrs, PhysicalIndex)->isr_TransmissionComplete!= NULL, ec_InvalidConfiguration); // no isr configured for this interrupt!
        break;
    case tit_USART_TransmitDataEmpty:
        CR1.TXEIE  = EnableIRQ;
        if (EnableIRQ)
            Assert_Interrupt(A(stm32_usart_isrs, PhysicalIndex)->isr_TransmitDataEmpty!= NULL, ec_InvalidConfiguration); // no isr configured for this interrupt!
        break;
    case tit_USART_RxNE:
        CR1.RXNEIE = EnableIRQ;
        if (EnableIRQ)
            Assert_Interrupt(A(stm32_usart_isrs, PhysicalIndex)->isr_ReceiveDataNotEmpty!= NULL, ec_InvalidConfiguration); // no isr configured for this interrupt!
        break;
    case tit_USART_LinBreak:
        CR2.LBDIE  = EnableIRQ;
        if (EnableIRQ)
            Assert_Interrupt(A(stm32_usart_isrs, PhysicalIndex)->isr_LinBreakDetected!= NULL, ec_InvalidConfiguration); // no isr configured for this interrupt!
        break;
    case tit_USART_Error:
        if (EnableIRQ)
            Assert_Interrupt(A(stm32_usart_isrs, PhysicalIndex)->isr_Error!= NULL, ec_InvalidConfiguration); // no isr configured for this interrupt!
        CR1.PEIE   = EnableIRQ;  // irq on parity error
        CR3.EIE    = EnableIRQ;  // irq on general error
        break;
    default: break;
    }

    if (! EnableIRQ) { // disable interrupt channel only if no other interrupt event for this USART is active
        if ( (CR1.PEIE   == 0) &&
             (CR1.IDLEIE == 0) &&
             (CR1.TCIE   == 0) &&
             (CR1.TXEIE  == 0) &&
             (CR1.RXNEIE == 0) &&
             (CR2.LBDIE  == 0) &&
             (CR3.CTSIE  == 0) &&
             (CR3.EIE    == 0)
             )
            NVIC_Cfg.NVIC_IRQChannelCmd = DISABLE;
    }
    NVIC_Init(&NVIC_Cfg);

    // Note: We have to copy struct-data in this odd way because USART_Base->CR1.TXEIE = 1; gives unpredictable results!
    ((USART_TypeDef*) USART_Base)->CR1 = *( (u16_t*) &CR1 );
    ((USART_TypeDef*) USART_Base)->CR2 = *( (u16_t*) &CR2 );
    ((USART_TypeDef*) USART_Base)->CR3 = *( (u16_t*) &CR3 );
}
void stm32_interrupt_enable_gpio(ttc_interrupt_type_e Type, physical_index_t PhysicalIndex, BOOL EnableIRQ) {

    GPIO_TypeDef* GPIOx;
    u8_t Pin;
    stm32_gpio_from_index8(PhysicalIndex, &GPIOx, &Pin);
#if (TTC_ASSERT_INTERRUPTS==1)
    Assert_Interrupt(GPIOx,    ec_InvalidArgument);
    Assert_Interrupt(Pin < 16, ec_InvalidArgument);
    Assert_Interrupt(!ttc_array_is_null(stm32_gpio_isrs), ec_InvalidConfiguration); // init interrupt first!
    stm32_gpio_isrs_t* GPIO_ISR_Entry = A(stm32_gpio_isrs, Pin);
    Assert_Interrupt(GPIO_ISR_Entry, ec_InvalidConfiguration); // init interrupt first!
#endif

    static EXTI_InitTypeDef* EXTI_Cfg = NULL;
    if (EXTI_Cfg == NULL) { // first call: allocate structure
        EXTI_Cfg = ttc_memory_alloc_zeroed(sizeof(EXTI_InitTypeDef));
        EXTI_StructInit(EXTI_Cfg);
        EXTI_Cfg->EXTI_Mode = EXTI_Mode_Interrupt;
    }

    switch (Type) { // select trigger according to Type

    case tit_GPIO_Falling:
        EXTI_Cfg->EXTI_Trigger = EXTI_Trigger_Falling;
        break;

    case tit_GPIO_Rising:
        EXTI_Cfg->EXTI_Trigger = EXTI_Trigger_Rising;
        break;

    case tit_GPIO_Rising_Falling:
        EXTI_Cfg->EXTI_Trigger = EXTI_Trigger_Rising_Falling;
        break;

    default:
        Assert_Interrupt(0,ec_InvalidArgument);
        break;
    }

    EXTI_Cfg->EXTI_Line    = 1 << Pin;
    EXTI_Cfg->EXTI_LineCmd = EnableIRQ;

    EXTI_Init(EXTI_Cfg);
}

//{ irq-handlers (supported) ---------------------------
typedef struct {
    unsigned MemoryManageFault            : 1; // Memory Manage Address Register (MMAR) contains valid fault addressing value
    unsigned BusFault                     : 1; // Bus Fault Address Register (BFAR) contains valid fault addressing value
    unsigned BusError_DataAccess          : 1; // Bus error occured during data access
    unsigned BusError_DataAccessImprecise : 1; // bus error during data access (no stack frame or fault address known)
    unsigned StackError                   : 1; // Error during stacking: Stack pointer corrupt/ Stack size too large (reaching forbidden region)
    unsigned UnStackError                 : 1; // Error during unstacking: Stack pointer corrupted during exception/ MPU configuration changed during exception
    unsigned DataAccessViolation          : 1; // user application tried to access privileged memory region
    unsigned InstructionAccessViolation   : 1; // user application tried to execute code in privileged memory region
    unsigned BusFaultDuringStacking       : 1; // a bus fault occured during a stacking operation (Stack pointer corrupt/ Stack overrun/ PSP used uninitialized)
    unsigned BusFaultDuringUnStacking     : 1; // a bus fault occured during a unstacking operation (Stack pointer corrupted during exception)
    unsigned InstructionBusError          : 1; // branch to invalid memory region
    unsigned DivisionByZeroError          : 1; // division by zero
    unsigned UnalignedDataAccess          : 1; // unaligned data access with UNALIGN_TRP set
    unsigned NoCoprocessorAvailable       : 1; // attempt to execute a coprocessor instruction
    unsigned InvalidProgramCounter        : 1; // compare "The definitive guid to the ARM Cortex-M3" p.427
    unsigned InvalidState                 : 1; // compare "The definitive guid to the ARM Cortex-M3" p.428
    unsigned UndefinedInstruction         : 1; // instruction found not known to CortexM3
    unsigned DebugFaultEvent              : 1; // fault is caused by a debug event
    unsigned ForcedEscalation             : 1; // hardfault was forced by escalation of fault that could not be handled
    unsigned VectorFetchFailed            : 1; // bus fault at vector fetch/ incorrect vector table offset
    unsigned reserved                     : 12;
} ttc_interrupt_faults_t;

void hard_fault_handler_c(unsigned int * hardfault_args) {

    // reconstruct registers
    unsigned int R0 =  hardfault_args[0];
    unsigned int R1 =  hardfault_args[1];
    unsigned int R2 =  ((unsigned int) hardfault_args[2]);
    unsigned int R3 =  ((unsigned int) hardfault_args[3]);
    unsigned int R12 = ((unsigned int) hardfault_args[4]);
    unsigned int LR =  ((unsigned int) hardfault_args[5]);
    unsigned int PC =  ((unsigned int) hardfault_args[6]);
    unsigned int PSR = ((unsigned int) hardfault_args[7]);

    SCB_t registerSCB;
    u8_t Words2Copy = sizeof(SCB_t) / 4;
    u32_t* Writer = (u32_t*) &registerSCB;
    u32_t* Reader = (u32_t*) SCB;
    while (Words2Copy > 0) {
        Words2Copy--;
        *Writer++ = *Reader++;
    }
    volatile SCB_CFSR_t   CFSR = registerSCB.CFSR;
    volatile SCB_HFSR_t   HFSR = registerSCB.HFSR;

    ttc_interrupt_faults_t Reasons;
    *( (u32_t*) (&Reasons) ) = 0;

    // determine address of fault
    volatile void* FaultingAddress = (void*) PC;

    if (CFSR.MMFSR.MMARVALID) {
        Reasons.MemoryManageFault = 1;
        FaultingAddress = (void*) registerSCB.MMFAR;
    }
    if (CFSR.MMFSR.MSTKERR) {
        Reasons.StackError = 1;
    }
    if (CFSR.MMFSR.MUNSTKERR) {
        Reasons.UnStackError = 1;
    }
    if (CFSR.MMFSR.DACCVIOL) {
        Reasons.DataAccessViolation = 1;
    }
    if (CFSR.MMFSR.IACCVIOL) {
        Reasons.InstructionAccessViolation = 1;
    }
    if (CFSR.BFSR.BFARVALID) {
        Reasons.BusFault= 1;
        FaultingAddress = (void*) registerSCB.BFAR;
    }
    if (CFSR.BFSR.STKERR) {
        Reasons.BusFaultDuringStacking = 1;
    }
    if (CFSR.BFSR.UNSTKERR) {
        Reasons.BusFaultDuringUnStacking = 1;
    }
    if (CFSR.BFSR.IMPRECISERR) {
        Reasons.BusError_DataAccessImprecise = 1;
    }
    if (CFSR.BFSR.PRECISEERR) {
        Reasons.BusError_DataAccess = 1;
    }
    if (CFSR.BFSR.IBUSERR) {
        Reasons.InstructionBusError = 1;
    }
    if (CFSR.UFSR.DIVBYZERO) {
        Reasons.DivisionByZeroError = 1;
    }
    if (CFSR.UFSR.UNALIGNED) {
        Reasons.UnalignedDataAccess = 1;
    }
    if (CFSR.UFSR.NOCP) {
        Reasons.NoCoprocessorAvailable = 1;
    }
    if (CFSR.UFSR.INVPC) {
        Reasons.InvalidProgramCounter = 1;
    }
    if (CFSR.UFSR.INVSTATE) {
        Reasons.InvalidState = 1;
    }
    if (CFSR.UFSR.UNDEFINSTR) {
        Reasons.UndefinedInstruction = 1;
    }
    if (HFSR.DebugVT) {
        Reasons.DebugFaultEvent = 1;
    }
    if (HFSR.Forced) {
        Reasons.ForcedEscalation = 1;
    }
    if (HFSR.VectorTable) {
        Reasons.VectorFetchFailed = 1;
    }

    ((void) (R0+R1+R2+R3+R12+LR+PC+PSR)); // unused vars

    // Hint - To see cause of hardfault type the following in your gdb:
    // l *PC
    // p Reasons
    // x FaultingAddress

    Assert_Interrupt(0, ec_HardFault);
}
void HardFault_Handler() {

    __asm__(".global HardFault_Handler");
    __asm__(".extern hard_fault_handler_c");
    __asm__("TST LR, #4");
    __asm__("ITE EQ");
    __asm__("MRSEQ R0, MSP");
    __asm__("MRSNE R0, PSP");
    __asm__("B hard_fault_handler_c");
}

void USART_General_IRQHandler(physical_index_t PhysicalIndex, volatile USART_t* USART) {
    USART_SR_t StatusRegister = USART->SR;  // reading from SR and then from DR will reset flags in SR

    stm32_usart_isrs_t* ISRs = A(stm32_usart_isrs, PhysicalIndex);
    Assert_Interrupt(ISRs != NULL, ec_UNKNOWN); // should never happen: the usart indexed by PhysicalIndex has not been initialized!

    if (StatusRegister.RXNE && ISRs->isr_ReceiveDataNotEmpty) {
        ISRs->isr_ReceiveDataNotEmpty(PhysicalIndex, ISRs->Argument_ReceiveDataNotEmpty);
    }
    if (StatusRegister.TXE  && ISRs->isr_TransmitDataEmpty) {
        ISRs->isr_TransmitDataEmpty(PhysicalIndex, ISRs->Argument_TransmitDataEmpty);
    }
    if (StatusRegister.CTS  && ISRs->isr_ClearToSend) {
        ISRs->isr_ClearToSend(PhysicalIndex, ISRs->Argument_ClearToSend);
    }
    if (StatusRegister.TC   && ISRs->isr_TransmissionComplete) {
        ISRs->isr_TransmissionComplete(PhysicalIndex, ISRs->Argument_TransmissionComplete);
    }
    if (StatusRegister.IDLE && ISRs->isr_IdleLine) {
        ISRs->isr_IdleLine(PhysicalIndex, ISRs->Argument_IdleLine);
    }
    if (StatusRegister.LBD  && ISRs->isr_LinBreakDetected) {
        ISRs->isr_LinBreakDetected(PhysicalIndex, ISRs->Argument_LinBreakDetected);
    }

    if (  ( (*( (u8_t*) &StatusRegister )) & 0xf) &&  // error occured: gather data + call error handler
          (ISRs->isr_Error != NULL)                   // error-handler has been defined
          ) {
        ttc_usart_Error_t Error; *((u8_t*) &Error) = 0;
        if (StatusRegister.FE)   Error.Framing = 1;
        if (StatusRegister.NE)   Error.Noise   = 1;
        if (StatusRegister.ORE)  Error.Overrun = 1;
        if (StatusRegister.PE)   Error.Parity  = 1;

        ISRs->isr_Error(PhysicalIndex, Error, ISRs->Argument_Error);
    }
}
void USART1_IRQHandler() {
    USART_General_IRQHandler(0, (USART_t*) USART1);
    USART1->SR = 0; // clear all interrupt flags
}
void USART2_IRQHandler() {
    USART_General_IRQHandler(1, (USART_t*) USART2);
    USART2->SR = 0; // clear all interrupt flags
}
void USART3_IRQHandler() {
    USART_General_IRQHandler(2, (USART_t*) USART3);
    USART3->SR = 0; // clear all interrupt flags
}
void UART4_IRQHandler() {
    USART_General_IRQHandler(3, (USART_t*) UART4);
    UART4->SR = 0; // clear all interrupt flags
}
void UART5_IRQHandler() {
    USART_General_IRQHandler(4, (USART_t*) UART5);
    UART5->SR = 0; // clear all interrupt flags
}

void GPIO_General_IRQHandler(u8_t InterruptLine) {
    stm32_gpio_isrs_t* Entry = A(stm32_gpio_isrs, InterruptLine);
    if (Entry->isr)
        Entry->isr(Entry->PhysicalIndex, Entry->Argument);
}
void EXTI0_IRQHandler() {
    GPIO_General_IRQHandler(0);
    EXTI_ClearITPendingBit(EXTI_Line0);
}
void EXTI1_IRQHandler() {
    GPIO_General_IRQHandler(1);
    EXTI_ClearITPendingBit(EXTI_Line1);
}
void EXTI2_IRQHandler() {
    GPIO_General_IRQHandler(2);
    EXTI_ClearITPendingBit(EXTI_Line2);
}
void EXTI3_IRQHandler() {
    GPIO_General_IRQHandler(3);
    EXTI_ClearITPendingBit(EXTI_Line3);
}
void EXTI4_IRQHandler() {
    GPIO_General_IRQHandler(4);
    EXTI_ClearITPendingBit(EXTI_Line4);
}
void EXTI9_5_IRQHandler() {

    if(EXTI_GetITStatus(EXTI_Line5)) {
        // call the general GPIO Interrupt handler
        GPIO_General_IRQHandler(5);
        EXTI_ClearITPendingBit(EXTI_Line5);

    } else if(EXTI_GetITStatus(EXTI_Line6)) {
        // call the general GPIO Interrupt handler
        GPIO_General_IRQHandler(6);
        EXTI_ClearITPendingBit(EXTI_Line6);

    } else if(EXTI_GetITStatus(EXTI_Line7)) {
        // call the general GPIO Interrupt handler
        GPIO_General_IRQHandler(7);
        EXTI_ClearITPendingBit(EXTI_Line7);

    } else if(EXTI_GetITStatus(EXTI_Line8)) {
        // call the general GPIO Interrupt handler
        GPIO_General_IRQHandler(8);
        EXTI_ClearITPendingBit(EXTI_Line8);

    } else if(EXTI_GetITStatus(EXTI_Line9)) {
        // call the general GPIO Interrupt handler
        GPIO_General_IRQHandler(9);
        EXTI_ClearITPendingBit(EXTI_Line9);

    } else{
        Assert_Interrupt(0, ec_UNKNOWN);
    }
}
void EXTI15_10_IRQHandler() {
    if(EXTI_GetITStatus(EXTI_Line10)) {
        // call the general GPIO Interrupt handler
        GPIO_General_IRQHandler(10);
        EXTI_ClearITPendingBit(EXTI_Line10);

    } else if(EXTI_GetITStatus(EXTI_Line11)) {
        // call the general GPIO Interrupt handler
        GPIO_General_IRQHandler(11);
        EXTI_ClearITPendingBit(EXTI_Line11);

    } else if(EXTI_GetITStatus(EXTI_Line12)) {
        // call the general GPIO Interrupt handler
        GPIO_General_IRQHandler(12);
        EXTI_ClearITPendingBit(EXTI_Line12);

    } else if(EXTI_GetITStatus(EXTI_Line13)) {
        // call the general GPIO Interrupt handler
        GPIO_General_IRQHandler(13);
        EXTI_ClearITPendingBit(EXTI_Line13);

    } else if(EXTI_GetITStatus(EXTI_Line14)) {
        // call the general GPIO Interrupt handler
        GPIO_General_IRQHandler(14);
        EXTI_ClearITPendingBit(EXTI_Line14);

    } else if(EXTI_GetITStatus(EXTI_Line15)) {
        // call the general GPIO Interrupt handler
        GPIO_General_IRQHandler(15);
        EXTI_ClearITPendingBit(EXTI_Line15);

    } else{
        Assert_Interrupt(0, ec_UNKNOWN);
    }
}

//}irq-handlers (supported)
//{ irq-handlers (yet unsupported) ---------------------------

/* void Reset_Handler() {
  // write your own reset-handler here...


  Assert_Interrupt(0, ec_UNKNOWN);
} */
void NMI_Handler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void MemManage_Handler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void BusFault_Handler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void UsageFault_Handler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void SVC_Handler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void DebugMon_Handler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void PendSV_Handler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void SysTick_Handler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void WWDG_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void PVD_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void TAMPER_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void RTC_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void FLASH_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void RCC_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void DMA1_Channel1_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void DMA1_Channel2_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void DMA1_Channel3_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void DMA1_Channel4_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void DMA1_Channel5_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void DMA1_Channel6_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void DMA1_Channel7_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void ADC1_2_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void CAN1_TX_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void CAN1_RX0_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void CAN1_RX1_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void CAN1_SCE_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void TIM1_BRK_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void TIM1_UP_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void TIM1_TRG_COM_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void TIM1_CC_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void TIM2_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void TIM3_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void TIM4_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void I2C1_EV_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void I2C1_ER_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void I2C2_EV_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void I2C2_ER_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void SPI1_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void SPI2_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void RTCAlarm_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void OTG_FS_WKUP_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void TIM5_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void SPI3_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void TIM6_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void TIM7_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void DMA2_Channel1_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void DMA2_Channel2_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void DMA2_Channel3_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void DMA2_Channel4_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void DMA2_Channel5_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void ETH_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void ETH_WKUP_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void CAN2_TX_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void CAN2_RX0_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void CAN2_RX1_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}
void CAN2_SCE_IRQHandler() {
    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}


#ifdef USE_USB_OTG_FS
void OTG_FS_IRQHandler(void)
#else
void OTG_HS_IRQHandler(void)
#endif
{
#ifdef EXTENSION_400_stm32_usb_fs_device_lib
    STM32_PCD_OTG_ISR_Handler();
#elif EXTENSION_400_stm32_usb_fs_host_lib
    USBH_OTG_ISR_Handler(&USB_OTG_Core_dev);
#else
    Assert_Interrupt(0, ec_UNKNOWN);
#endif
}

//#ifdef STM32F10X_CL
//void OTG_FS_IRQHandler() {
//    // call your interrupt-handler right here, or else ...
//#ifdef EXTENSION_400_stm32_usb_fs_device_lib
//    STM32_PCD_OTG_ISR_Handler();
//#elif EXTENSION_400_stm32_usb_fs_host_lib
//    USBH_OTG_ISR_Handler(&USB_OTG_Core_dev);
//#else
//    Assert_Interrupt(0, ec_UNKNOWN);
//#endif
//}
//#endif
void BootRAM() {

    // call your interrupt-handler right here, or else ...


    Assert_Interrupt(0, ec_UNKNOWN);
}

//} irq-handlers (yet unsupported)

//} Function definitions
