/*{ ttc_queue.c ***********************************************
 *
 * Written by Gregor Rebel 2010-2012
 *
 * Architecture independent support for inter task communication & synchronization.
 * 
 * Currently supported schedulers: FreeRTOS
}*/
#include "ttc_queue.h"

#define AMOUNT_LAST_SMART_MUTEXES 0
#if AMOUNT_LAST_SMART_MUTEXES > 0

    // stores pointers to smart mutexes given for latest ttc_mutex_lock_smart() calls
    ttc_mutex_smart_t* tml_LastSmartMutex[AMOUNT_LAST_SMART_MUTEXES];

    // history of smart mutexes without doubles
    ttc_mutex_smart_t* tml_UniqueSmartMutex[AMOUNT_LAST_SMART_MUTEXES];

    // index of next valid entry inside tml_LastSmartMutex[]
    u8_t tml_Index_LastSmartMutex = 0;
    u8_t tml_Index_UniqueSmartMutex = 0;

    u32_t tml_AmountSmartLocks = 0;
#endif

//{ functions *************************************************

ttc_QueueHandle ttc_queue_create(Base_t AmountOfEntries, Base_t SizeOfEntries) {
    ttc_QueueHandle NewQueue = NULL;

#ifdef EXTENSION_300_scheduler_free_rtos
    NewQueue = (ttc_QueueHandle) xQueueCreate(AmountOfEntries, SizeOfEntries);
#endif

    Assert(NewQueue != NULL, ec_Malloc); // could not create queue (no implementation/ not egnough memory)
    ttc_task_check_stack(); //D

    return NewQueue;
}
ttc_queue_error_e ttc_queue_push_back(ttc_QueueHandle Queue, const void* Entry, Base_t TimeOut) {
    Assert_QUEUES(Queue != 0, ec_InvalidArgument);
    Assert_QUEUES(Entry != 0, ec_InvalidArgument);

    ttc_task_check_stack(); //D
#ifdef EXTENSION_300_scheduler_free_rtos
    if (xQueueSend( (xQueueHandle) Queue, Entry, TimeOut ) == pdTRUE)
        return tqe_OK;
    else
        return tqe_TimeOut;
#endif

    return tqe_NotImplemented;
}
ttc_queue_error_e ttc_queue_push_back_isr(ttc_QueueHandle Queue, const void* Entry) {
    Assert_QUEUES(Queue != 0, ec_InvalidArgument);
    Assert_QUEUES(Entry != 0, ec_InvalidArgument);

#ifdef EXTENSION_300_scheduler_free_rtos
    portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;
    signed portBASE_TYPE Status = xQueueSendToBackFromISR( (xQueueHandle) Queue, Entry, &xHigherPriorityTaskWoken);

    if (xHigherPriorityTaskWoken == pdFALSE) {
        if (Status == pdPASS)
            return tqe_OK;
        else
            return tqe_QueuePushFailed;
    }
    else
        return tqe_HigherPriorityTaskWoken;
#endif

    return tqe_NotImplemented;
}
ttc_queue_error_e ttc_queue_pop_front(ttc_QueueHandle Queue, void* Entry, Base_t TimeOut) {
    Assert_QUEUES(Queue != 0, ec_InvalidArgument);
    Assert_QUEUES(Entry != 0, ec_InvalidArgument);

#ifdef EXTENSION_300_scheduler_free_rtos
    if ( xQueueReceive( (xQueueHandle) Queue, Entry, TimeOut ) )
        return tqe_OK;
    else
        return tqe_TimeOut;
#endif

    return tqe_NotImplemented;
}
ttc_queue_error_e ttc_queue_pop_front_isr(ttc_QueueHandle Queue, void* Entry) {
    Assert_QUEUES(Queue != 0, ec_InvalidArgument);
    Assert_QUEUES(Entry != 0, ec_InvalidArgument);

#ifdef EXTENSION_300_scheduler_free_rtos
    signed portBASE_TYPE pxTaskWoken = 0;
    if ( xQueueReceiveFromISR( (xQueueHandle) Queue, Entry, &pxTaskWoken) ) {
        if (pxTaskWoken)
            return tqe_HigherPriorityTaskWoken;
        return tqe_OK;
    }
    else
        return tqe_TimeOut;
#endif

    return tqe_NotImplemented;
}
Base_t ttc_queue_amount_waiting(ttc_QueueHandle Queue) {
    Assert_QUEUES(Queue != 0, ec_InvalidArgument);

#ifdef EXTENSION_300_scheduler_free_rtos
    return uxQueueMessagesWaiting( (xQueueHandle) Queue );
#endif

    return 0;
}
ttc_MutexHandle ttc_mutex_create() {

    ttc_MutexHandle NewMutex = NULL;

#ifdef EXTENSION_300_scheduler_free_rtos
    NewMutex = xSemaphoreCreateMutex();
#endif
    Assert_QUEUES(NewMutex, ec_NULL);

    return NewMutex;
}
ttc_mutex_smart_t* ttc_mutex_create_smart() {

    ttc_mutex_smart_t* NewMutex = ttc_memory_alloc( sizeof(ttc_mutex_smart_t) );
    NewMutex->Mutex = ttc_mutex_create();

    return NewMutex;
}

ttc_queue_error_e ttc_mutex_lock_smart(ttc_mutex_smart_t* SmartMutex, Base_t TimeOut, void (*Owner)() ) {
    Assert_QUEUES(SmartMutex, ec_InvalidArgument);

#if AMOUNT_LAST_SMART_MUTEXES > 0
    tml_AmountSmartLocks++;
    tml_LastSmartMutex[tml_Index_LastSmartMutex++] = SmartMutex;
    if (tml_Index_LastSmartMutex >= AMOUNT_LAST_SMART_MUTEXES)
        tml_Index_LastSmartMutex = 0;

    BOOL MutexIsKnown = FALSE;
    for (int I = 0; (I < tml_Index_UniqueSmartMutex) && (I < AMOUNT_LAST_SMART_MUTEXES); I++) {
        if (tml_UniqueSmartMutex[I] == SmartMutex) {
            MutexIsKnown = TRUE;
            break;
        }
    }
    if ( (!MutexIsKnown) && (tml_Index_UniqueSmartMutex < AMOUNT_LAST_SMART_MUTEXES) )
        tml_UniqueSmartMutex[tml_Index_UniqueSmartMutex++] = SmartMutex;

#endif
    //TimeOut = 2000000; //D

    ttc_queue_error_e Error = ttc_mutex_lock(SmartMutex->Mutex, TimeOut);
    if (Error == tqe_OK)
        SmartMutex->Owner = Owner;
    //Assert(Error==tqe_OK, ec_TimeOut); //D

    return Error;
}
u32_t tml_Status = 0; //D
u32_t tml_CountTimedLocks = 0; //D

ttc_queue_error_e ttc_mutex_lock(ttc_MutexHandle Mutex, Base_t TimeOut) {
    Assert_QUEUES(Mutex != 0, ec_InvalidArgument);

#ifdef EXTENSION_300_scheduler_free_rtos

    // offset used to calibrate timeouts for current scheduler
    static Base_t TimeOffset = 0;

    if (TimeOut != -1) {
        if (TimeOut > TimeOffset)
            TimeOut -= TimeOffset;
        else
            TimeOut = 0;

        portTickType Ticks2Wait = (portTickType) TimeOut / (1000000 / configTICK_RATE_HZ);

        if (TimeOffset == 0) { // calibrate mutex
tml_Status=4; //D
            Base_t StartTime = ttc_task_get_tick_count();
            u8_t Result = xSemaphoreTake(Mutex, Ticks2Wait);
            if (!Result) { // timeout: measure overhead
                Base_t EndTime = ttc_task_get_tick_count();
                TimeOffset = EndTime - StartTime - TimeOut;
                tml_Status=5; //D
                return tqe_TimeOut;
            }
            tml_Status=6; //D
            return tqe_OK;
        }
        else {
            tml_Status=1; //D
            tml_CountTimedLocks++; //D
            if (xSemaphoreTake(Mutex, Ticks2Wait)) {
                tml_Status=2; //D
                return tqe_OK;
            }
            else {
                tml_Status=3; //D
                return tqe_TimeOut;
            }
        }
    }
    else { // wait forever
tml_Status=7; //D
        if (0)
        { while (xSemaphoreTake(Mutex, (portTickType) -1) == 0); }
        else { // easier to debug: will call ttc_task_yield() periodically while waiting
            while (xSemaphoreTake(Mutex, (portTickType) 1000) == 0)
                ttc_mutex_wait();
        }
tml_Status=8; //D
        return tqe_OK;
    }
#endif

    return tqe_NotImplemented;
}
void ttc_mutex_wait() {
    ttc_task_yield();
}
ttc_queue_error_e ttc_mutex_lock_isr(ttc_MutexHandle Mutex, BOOL* HigherPrioTaskWoken) {
    Assert_QUEUES(Mutex != 0, ec_InvalidArgument);

#ifdef EXTENSION_300_scheduler_free_rtos
    signed portBASE_TYPE Return;

    if (HigherPrioTaskWoken) {
        signed portBASE_TYPE TaskWoken = 0;
        Return = xSemaphoreTakeFromISR(Mutex, &TaskWoken);
        if (TaskWoken)
            *HigherPrioTaskWoken = TRUE;
        else
            *HigherPrioTaskWoken = FALSE;
    }
    else {
        Return = xSemaphoreTakeFromISR(Mutex, NULL);
    }

    if (Return)
        return tqe_OK;
    else
        return tqe_MutexLocked;
#endif

    return tqe_NotImplemented;
}
ttc_queue_error_e ttc_mutex_unlock(ttc_MutexHandle Mutex) {
    Assert_QUEUES(Mutex != 0, ec_InvalidArgument);

#ifdef EXTENSION_300_scheduler_free_rtos
    if ( xSemaphoreGive(Mutex) )
        return tqe_OK;
    else
        return tqe_UNKNOWN;
#endif

    return tqe_NotImplemented;
}
ttc_queue_error_e _ttc_mutex_unlock_smart(ttc_mutex_smart_t* SmartMutex) {
    Assert_QUEUES(SmartMutex, ec_InvalidArgument);

    if (0) SmartMutex->Owner = NULL;

    return ttc_mutex_unlock(SmartMutex->Mutex);
}
ttc_queue_error_e ttc_mutex_unlock_isr(ttc_MutexHandle Mutex) {
    Assert_QUEUES(Mutex != 0, ec_InvalidArgument);

#ifdef EXTENSION_300_scheduler_free_rtos
    portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;
    xSemaphoreGiveFromISR(Mutex, &xHigherPriorityTaskWoken);

    if (xHigherPriorityTaskWoken == pdFALSE)
        return tqe_OK;
    else
        return tqe_HigherPriorityTaskWoken;
#endif

    return tqe_NotImplemented;
}
ttc_SemaphoreHandle ttc_semaphore_create(u8_t MaxCount, u8_t InitialCount) {

#ifdef EXTENSION_300_scheduler_free_rtos
        return xQueueCreateCountingSemaphore(MaxCount, InitialCount);
#endif

    return NULL;
}
ttc_queue_error_e ttc_semaphore_take(ttc_SemaphoreHandle Semaphore, Base_t TimeOut) {
    Assert_QUEUES(Semaphore != 0, ec_InvalidArgument);

#ifdef EXTENSION_300_scheduler_free_rtos
    if (xSemaphoreTake(Semaphore, (portTickType) TimeOut))
        return tqe_OK;
    else
        return tqe_TimeOut;
#endif

    return tqe_NotImplemented;
}
ttc_queue_error_e ttc_semaphore_give(ttc_SemaphoreHandle Semaphore) {
    Assert_QUEUES(Semaphore != 0, ec_InvalidArgument);

#ifdef EXTENSION_300_scheduler_free_rtos
    if ( xSemaphoreGive(Semaphore) )
        return tqe_OK;
    else
        return tqe_UNKNOWN;
#endif

    return tqe_NotImplemented;
}
ttc_queue_error_e ttc_semaphore_give_isr(ttc_SemaphoreHandle Semaphore) {
    Assert_QUEUES(Semaphore != 0, ec_InvalidArgument);

#ifdef EXTENSION_300_scheduler_free_rtos
    portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;
    xSemaphoreGiveFromISR(Semaphore, &xHigherPriorityTaskWoken);

    if (xHigherPriorityTaskWoken == pdFALSE)
        return tqe_OK;
    else
        return tqe_HigherPriorityTaskWoken;
#endif

    return tqe_NotImplemented;
}
ttc_QueueHandle_Bytes_t ttc_queue_byte_create(u16_t AmountOfEntries) {
    Assert(AmountOfEntries, ec_InvalidArgument); // need at least size 1

    ttc_SimpleQueueBytes_t* Queue = ttc_memory_alloc(AmountOfEntries + sizeof(ttc_SimpleQueueBytes_t));
    Queue->Size  = AmountOfEntries;
    Queue->Start = 0;
    Queue->End   = 0;
    Queue->Lock = ttc_mutex_create();

    return Queue;
}
Base_t ttc_queue_byte_get_amount(ttc_QueueHandle_Bytes_t QueueHandle) {
    ttc_SimpleQueueBytes_t* Queue = (ttc_SimpleQueueBytes_t*) QueueHandle;
    Assert(Queue, ec_InvalidArgument);

    if (Queue->End < Queue->Start)
        return Queue->End + Queue->Size - Queue->Start;

    return Queue->End - Queue->Start;
}
ttc_queue_error_e ttc_queue_byte_push_back(ttc_QueueHandle_Bytes_t QueueHandle, u8_t Data) {
    ttc_SimpleQueueBytes_t* Queue = (ttc_SimpleQueueBytes_t*) QueueHandle;
    Assert(Queue, ec_InvalidArgument);

    ttc_mutex_lock(Queue->Lock, -1); // ToDo: Lock_ISR
    u16_t NewEnd = Queue->End + 1;
    if (NewEnd >= Queue->Size)
        NewEnd = 0;

    if (Queue->Start == NewEnd) {
        ttc_mutex_unlock(Queue->Lock);
        return tqe_QueuePushFailed; // queue is full
    }

    Queue->Data[Queue->End] = Data; // storing at old end index
    Queue->End = NewEnd;

    ttc_mutex_unlock(Queue->Lock);
    return tqe_OK;
}
ttc_queue_error_e ttc_queue_byte_pull_front(ttc_QueueHandle_Bytes_t QueueHandle, u8_t* Data) {
    ttc_SimpleQueueBytes_t* Queue = (ttc_SimpleQueueBytes_t*) QueueHandle;
    Assert(Queue, ec_InvalidArgument);
    Assert(Data,  ec_InvalidArgument);

    ttc_mutex_lock(Queue->Lock, -1); // ToDo: Lock_ISR
    u16_t Start = Queue->Start;
    Assert(Start < Queue->Size,  ec_UNKNOWN);

    if (Start == Queue->End) {      // queue is empty
        ttc_mutex_unlock(Queue->Lock);
        return tqe_QueueIsEmpty;
    }

    u8_t Byte = Queue->Data[Start++];
    *Data = Byte;

    if (Start == Queue->Size)
        Start = 0;                 // wrapping around at end of array
    Queue->Start = Start;

    ttc_mutex_unlock(Queue->Lock);
    return tqe_OK;
}
ttc_QueueHandle_Pointers_t ttc_queue_pointer_create(u16_t AmountOfEntries) {
    Assert(AmountOfEntries, ec_InvalidArgument); // need at least size 1

    ttc_SimpleQueuePointers_t* Queue = ttc_memory_alloc(sizeof(void*) * AmountOfEntries + sizeof(ttc_SimpleQueuePointers_t));
    Queue->Size  = AmountOfEntries;
    Queue->Start = 0;
    Queue->End   = 0;
    Queue->Lock  = ttc_mutex_create();

    return Queue;
}
Base_t ttc_queue_pointer_get_amount(ttc_QueueHandle_Pointers_t QueueHandle) {
    ttc_SimpleQueuePointers_t* Queue = (ttc_SimpleQueuePointers_t*) QueueHandle;
    Assert(Queue, ec_InvalidArgument);

    if (Queue->End < Queue->Start)
        return Queue->End + Queue->Size - Queue->Start;

    return Queue->End - Queue->Start;
}
ttc_queue_error_e ttc_queue_pointer_push_back(ttc_QueueHandle_Pointers_t QueueHandle, void* Pointer) {
    ttc_SimpleQueuePointers_t* Queue = (ttc_SimpleQueuePointers_t*) QueueHandle;
    Assert(Queue, ec_InvalidArgument);

    if (1) ttc_mutex_lock(Queue->Lock, -1);

    Base_t NewEnd = Queue->End + 1;
    if (NewEnd >= Queue->Size)
        NewEnd = 0;                 // wrapping around at end of array

    if (Queue->Start == NewEnd) {
        if (1) ttc_mutex_unlock(Queue->Lock);
        return tqe_QueuePushFailed; // queue is full
    }

    Queue->Data[Queue->End] = Pointer; // storing at old end index
    Queue->End = NewEnd;

    if (1) ttc_mutex_unlock(Queue->Lock);
    return tqe_OK;
}
ttc_queue_error_e ttc_queue_pointer_pull_front(ttc_QueueHandle_Pointers_t QueueHandle, void** Pointer) {
    ttc_SimpleQueuePointers_t* Queue = (ttc_SimpleQueuePointers_t*) QueueHandle;
    Assert(Queue,   ec_InvalidArgument);
    Assert(Pointer, ec_InvalidArgument);

    if (1) ttc_mutex_lock(Queue->Lock, -1);

    u16_t Start = Queue->Start;
    Assert(Start < Queue->Size,  ec_UNKNOWN);

    if (Start == Queue->End) {   // queue is empty
        *Pointer = NULL;
        if (1) ttc_mutex_unlock(Queue->Lock);
        return tqe_QueueIsEmpty;
    }
    void* Address = Queue->Data[Start++];
    *Pointer = Address;

    if (Start == Queue->Size)
        Start = 0;              // wrapping around at end of array
    Queue->Start = Start;

    if (1) ttc_mutex_unlock(Queue->Lock);
    return tqe_OK;
}

//}functions
