#ifndef ttc_queue_H
#define ttc_queue_H

/*{ ttc_queue.h ***********************************************
 * 
 * Written by Gregor Rebel 2010-2012
 *
 * Architecture independent support for inter task communication & synchronization.
 * 
 * Currently supported schedulers: FreeRTOS
 * 
}*/

//{ includes

#include "ttc_basic.h"
#include "ttc_memory.h"
#include "ttc_task.h" //D

#ifdef EXTENSION_300_scheduler_free_rtos
  #include <string.h>
  #include "stddef.h"
  #include "FreeRTOS.h"
  #include "queue.h"
  #include "semphr.h"

typedef struct QueueDefinition { // copied from queue.c (use only for debugging purpose)
    signed char *pcHead;				/*< Points to the beginning of the queue storage area. */
    signed char *pcTail;				/*< Points to the byte at the end of the queue storage area.  Once more byte is allocated than necessary to store the queue items, this is used as a marker. */

    signed char *pcWriteTo;				/*< Points to the free next place in the storage area. */
    signed char *pcReadFrom;			/*< Points to the last place that a queued item was read from. */

    xList xTasksWaitingToSend;				/*< List of tasks that are blocked waiting to post onto this queue.  Stored in priority order. */
    xList xTasksWaitingToReceive;			/*< List of tasks that are blocked waiting to read from this queue.  Stored in priority order. */

    volatile unsigned portBASE_TYPE uxMessagesWaiting;/*< The number of items currently in the queue. */
    unsigned portBASE_TYPE uxLength;		/*< The length of the queue defined as the number of items it will hold, not the number of bytes. */
    unsigned portBASE_TYPE uxItemSize;		/*< The size of each items that the queue will hold. */

    signed portBASE_TYPE xRxLock;			/*< Stores the number of items received from the queue (removed from the queue) while the queue was locked.  Set to queueUNLOCKED when the queue is not locked. */
    signed portBASE_TYPE xTxLock;			/*< Stores the number of items transmitted to the queue (added to the queue) while the queue was locked.  Set to queueUNLOCKED when the queue is not locked. */

    #if ( configUSE_TRACE_FACILITY == 1 )
        unsigned char ucQueueNumber;
        unsigned char ucQueueType;
    #endif

} xQUEUE_Structure;

// generic queue handle provided by scheduler (E.g. FreeRTOS)
typedef xQUEUE_Structure* ttc_QueueHandle;
#else

typedef void* ttc_QueueHandle;

#endif

typedef enum ttc_queue_error_E {      // return type of most functions in ttc_queue
    tqe_OK,                           // no error
    tqe_TimeOut,                      // an operation has timed out
    tqe_NotImplemented,               // required function not implemented for current architecture
    tqe_HigherPriorityTaskWoken,      // an _isr()-call has unblocked a higher priority task to unblock. Call ttc_task_yield() before exiting your ISR!
    tqe_QueuePushFailed,              // pushing data to a queue has failed
    tqe_QueueIsEmpty,                 // pulling data from a queue has failed because its empty
    tqe_MutexLocked,                  // tried to lock a mutex that is already locked
    tqe_UNKNOWN
} ttc_queue_error_e;

//}includes
//{ Default settings (change via makefile) ******************************

#ifndef TTC_ASSERT_QUEUES
  #define TTC_ASSERT_QUEUES 1
#endif

#ifndef TTC_SMART_MUTEXES
  #define TTC_SMART_MUTEXES 1
#endif

//}Default settings
//{ Types & Structures ***************************************************


#if TTC_ASSERT_QUEUES == 1 // use Assert()s in Queue code (somewhat slower but alot easier to debug)
  #define Assert_QUEUES(Condition, ErrorCode) Assert(Condition, ErrorCode)
#else  // us no Assert()s in USART code (somewhat smaller + faster, but crashes are hard to debug)
  #define Assert_QUEUES(Condition, ErrorCode)
#endif

// generic semaphore handle provided by scheduler (E.g. FreeRTOS)
typedef void* ttc_SemaphoreHandle;

// generic mutex handle provided by scheduler (E.g. FreeRTOS)
typedef void* ttc_MutexHandle;

// debuggable mutex that can store its owner
// This comes handy if you wanna know, who has locked a mutex before.
#if TTC_SMART_MUTEXES == 1
  typedef struct {
    ttc_MutexHandle Mutex;
    void (*Owner)();
  } ttc_mutex_smart_t;
#else
  typedef ttc_MutexHandle ttc_mutex_smart_t;
#endif

typedef struct { // data of a simple queue for 8-bit values
    ttc_MutexHandle Lock; // protects queue from concurrent accesses
    Base_t Start;   // index of first valid element in Data[]
    Base_t End;     // index of last valid element in Data[]
    Base_t Size;    // size of Data[]
    u8_t Data[];    // stores 8-bit values
} __attribute__((__packed__)) ttc_SimpleQueueBytes_t;
typedef struct { // data of a simple queue for pointer values
   ttc_MutexHandle Lock; // protects queue from concurrent accesses
   Base_t Start;   // index of first valid element in Data[]
   Base_t End;     // index of last valid element in Data[]
   Base_t Size;    // size of Data[]
    void* Data[]; // stores pointers
} __attribute__((__packed__)) ttc_SimpleQueuePointers_t;

// generic queue handle provided by scheduler (E.g. FreeRTOS)
typedef ttc_SimpleQueuePointers_t* ttc_QueueHandle_Pointers_t;

// generic queue handle provided by scheduler (E.g. FreeRTOS)
typedef ttc_SimpleQueueBytes_t*    ttc_QueueHandle_Bytes_t;

//}
//{ Function prototypes **************************************************

/** creates a new queue
 * @param AmountOfEntries  maximum amount of entries that the new queue can store
 * @param SizeOfEntries    size of each entry in queue
 * @return                 =! NULL: queue has been created successfully (use handle to operate on queue)
 */
ttc_QueueHandle ttc_queue_create(Base_t AmountOfEntries, Base_t SizeOfEntries);

/** adds given entry ad end of given queue
 * @param Queue    handle of already created queue
 * @param Entry    pointer to data to be copied into new queue entry
 * @param TimeOut  amount of system ticks to wait for queue to become available
 * @return         == tqe_OK: Entry has been successfully added to Queue within TimeOut
 */
ttc_queue_error_e ttc_queue_push_back(ttc_QueueHandle Queue, const void* Entry, Base_t TimeOut);

/** adds given entry ad end of given queue (does never block)
 * Note: Must be called from ISR or with disabled scheduler! (-> ttc_task_begin_criticalsection )
 *
 * @param Queue    handle of already created queue
 * @param Entry    pointer to data to be copied into new queue entry
 * @param TimeOut  amount of system ticks to wait for queue to become available
 * @return         == tqe_OK: Entry has been successfully added to Queue within TimeOut
 *                 == tqe_HigherPriorityTaskWoken: call task ttc_task_yield_isr() before exiting your ISR!
 */
ttc_queue_error_e ttc_queue_push_back_isr(ttc_QueueHandle Queue, const void* Entry);

/** removes first entry from given queue and copies its data into given buffer
 * @param Queue  handle of already created queue
 * @param Entry  pointer to buffer where content of first entry should be copied to
 * @param TimeOut  amount of system ticks to wait for queue to provide data
 * @return         == tqe_OK: Entry has been successfully added to Queue within TimeOut
 */
ttc_queue_error_e ttc_queue_pop_front(ttc_QueueHandle Queue, void* Entry, Base_t TimeOut);

/** removes first entry from given queue and copies its data into given buffer (does never block)
 * @param Queue  handle of already created queue
 * @param Entry  pointer to buffer where content of first entry should be copied to
 * @return         == tqe_OK: Entry has been successfully added to Queue within TimeOut
 */
ttc_queue_error_e ttc_queue_pop_front_isr(ttc_QueueHandle Queue, void* Entry);

/** provides amount of entries currently waiting in given queue
 * @param Queue  handle of already created queue
 * @return       current amount of occupied entries
 */
Base_t ttc_queue_amount_waiting(ttc_QueueHandle Queue);

/** creates a new non-recursive mutex lock
 * @return                 =! NULL: mutex has been created successfully (use handle to operate on mutex)
 */
ttc_MutexHandle ttc_mutex_create();

/** creates a new non-recursive mutex lock
 *
 * Smart mutexes store the owner of a lock.
 * Whenever a mutex stays locked because of a programming error, then the last locking
 * function can be identified by taking a look into the smart lock structure.
 *
 * @return                 =! NULL: mutex has been created successfully (use handle to operate on mutex)
 */
ttc_mutex_smart_t* ttc_mutex_create_smart();

/** Tries to obtain lock on mutex within timeout
 * @param Mutex  handle of already created mutex
 * @param TimeOut  !=-1: amount of system ticks to wait for lock
 *                 ==-1: wait forever
 * @return         == tqe_OK: Lock has been successfully obtained within TimeOut
 */
ttc_queue_error_e ttc_mutex_lock(ttc_MutexHandle Mutex, Base_t TimeOut);

/** Tries to obtain lock on mutex; this function will never block
 * @param Mutex                handle of already created mutex
 * @param HigherPrioTaskWoken  loaded with new value: ==TRUE - a task of higher priority has been interrupted => call ttc_task_yield() at end of your interrupt service routine!
 * @return         == tqe_OK: Lock has been successfully obtained within TimeOut
 */
ttc_queue_error_e ttc_mutex_lock_isr(ttc_MutexHandle Mutex, BOOL* HigherPrioTaskWoken);

/** Tries to obtain lock on a smart mutex within timeout
 *
 * Smart mutexes store the owner of a lock.
 * Whenever a mutex stays locked because of a programming error, then the last locking
 * function can be identified by taking a look into the smart lock structure.
 *
 * @param Mutex    handle of already created mutex
 * @param TimeOut  amount of system ticks to wait for lock
 * @param Owner    reference to function which wants to obtain the lock
 * @return         == tqe_OK: Lock has been successfully obtained within TimeOut
 */
ttc_queue_error_e ttc_mutex_lock_smart(ttc_mutex_smart_t* SmartMutex, Base_t TimeOut, void (*Owner)() );

/** Releases an already obtained lock from mutex
 * @param Mutex  handle of already created smart mutex
 * @return         == tqe_OK: Lock has been released successfully
 */
ttc_queue_error_e ttc_mutex_unlock(ttc_MutexHandle Mutex);
ttc_queue_error_e _ttc_mutex_unlock_smart(ttc_mutex_smart_t* SmartMutex);
#define ttc_mutex_unlock_smart(SmartMutex) _ttc_mutex_unlock_smart(SmartMutex)

/** Releases an already obtained lock from mutex (does never block)
 * @param Mutex  handle of already created mutex
 * @return         == tqe_OK: Lock has been released successfully
 *                   == tqe_HigherPriorityTaskWoken: call task ttc_task_yield_isr() before exiting your ISR!
 */
ttc_queue_error_e ttc_mutex_unlock_isr(ttc_MutexHandle Mutex);
#define ttc_mutex_unlock_isr_smart(SmartMutex) _ttc_mutex_unlock_smart(SmartMutex)

/** called while waiting for mutex lock. Place your breakpoint inside to debug hanging locks
  */
void ttc_mutex_wait();

/** creates a new counting semaphore
 * @param   MaxCount       maximum allowed locks for this semaphore
 * @param   InitialCount   initial value (semaphore can be unlocked InitialCount times)
 * @return                 =! NULL: semaphore has been created successfully (use handle to operate on mutex)
 */
ttc_SemaphoreHandle ttc_semaphore_create(u8_t MaxCount, u8_t InitialCount);

/** Tries to add one token to semaphore
 * @param Semaphore  handle of already created semaphore
 * @return           == tqe_OK: Token has been added successfully
 *                   == tqe_HigherPriorityTaskWoken: call task ttc_task_yield_isr() before exiting your ISR!
 */
ttc_queue_error_e ttc_semaphore_give(ttc_SemaphoreHandle Semaphore);

/** Tries to add one token to semaphore (does never block)
 * @param Semaphore  handle of already created semaphore
 * @return           == tqe_OK: Token has been added successfully
 */
ttc_queue_error_e ttc_semaphore_give_isr(ttc_SemaphoreHandle Semaphore);

/** Tries to obtaine one token from semaphore within timeout
 * @param Semaphore  handle of already created semaphore
 * @param TimeOut    amount of system ticks to wait for lock
 * @return           == tqe_OK: Token has been obtained successfully within TimeOut
 */
ttc_queue_error_e ttc_semaphore_take(ttc_SemaphoreHandle Semaphore, Base_t TimeOut);

/** Creates a simple queue for 8-bit data.
 * Simple queues are faster and use less memory than general queues.
 * Simple queues are available even without a multitasking scheduler.
 * Functions on simple queues will never block.
 *
 * Restrictions are:
 * - only one task is alowed to push data into the queue
 * - only one task is alowed to pop data from the queue
 * - fixed item size
 *
 * @param AmountOfEntries  Queue is created with a fixed size. It cannot be resized later.
 * @return pointer to struct containing all necessary data of this queue
 */
ttc_QueueHandle_Bytes_t ttc_queue_byte_create(u16_t AmountOfEntries);

/** returns amount of entries currently stored in given queue
 *
 * Note: This function is not protected against concurrent calls from different tasks!
 * Note: Returned size can be corrupt if items are added/ removed by another task!
 *
 * @param QueueHandle  handle of already created queue
 * @param Data         Byte to append to queue
 * @return             == tqe_OK: Entry has been successfully added
 */
Base_t ttc_queue_byte_get_amount(ttc_QueueHandle_Bytes_t QueueHandle);

/** adds given byte at end of given queue
 *
 * Note: This function is not protected against concurrent calls from different tasks!
 * Note: It is safe to have one single task call push and another single task call pull on the same queue.
 * Note: It is not safe to have >1 task push/ pull on same queue!
 *
 * @param QueueHandle  handle of already created queue
 * @param Data         Byte to append to queue
 * @return             == tqe_OK: Entry has been successfully added
 */
ttc_queue_error_e ttc_queue_byte_push_back(ttc_QueueHandle_Bytes_t QueueHandle, u8_t Data);

/** reads one byte from given queue
 *
 * Note: This function is not protected against concurrent calls from different tasks!
 * Note: It is safe to have one single task call push and another single task call pull on the same queue.
 * Note: It is not safe to have >1 task push/ pull on same queue!
 *
 * @param QueueHandle  handle of already created queue
 * @param Data         byte will be written to this address (no write in case of error)
 * @return             == tqe_OK: Entry has been successfully pulled
 */
ttc_queue_error_e ttc_queue_byte_pull_front(ttc_QueueHandle_Bytes_t QueueHandle, u8_t* Data);

/** Creates a simple queue to store pointers.
 * Simple queues are faster and use less memory than general queues.
 * Simple queues are available even without a multitasking scheduler.
 * Functions on simple queues will never block.
 *
 * Restrictions are:
 * - only one task is alowed to push data into the queue
 * - only one task is alowed to pop data from the queue
 * - fixed item size
 *
 * @param AmountOfEntries  Queue is created with a fixed size. It cannot be resized later.
 * @return pointer to struct containing all necessary data of this queue
 */
ttc_QueueHandle_Pointers_t ttc_queue_pointer_create(u16_t AmountOfEntries);

/** returns amount of entries currently stored in given queue
 *
 * Note: This function is not protected against concurrent calls from different tasks!
 * Note: Returned size can be corrupt if items are added/ removed by another task!
 *
 * @param QueueHandle  handle of already created queue
 * @param Data         Byte to append to queue
 * @return             == tqe_OK: Entry has been successfully added
 */
Base_t ttc_queue_pointer_get_amount(ttc_QueueHandle_Pointers_t QueueHandle);

/** adds given byte at end of given queue
 *
 * Note: This function is not protected against concurrent calls from different tasks!
 * Note: It is safe to have one single task call push and another single task call pull on the same queue.
 * Note: It is not safe to have >1 task push/ pull on same queue!
 *
 * @param QueueHandle  handle of already created queue
 * @param Data         pointer address to append to queue
 * @return             == tqe_OK: Entry has been successfully added
 */
ttc_queue_error_e ttc_queue_pointer_push_back(ttc_QueueHandle_Pointers_t QueueHandle, void* Pointer);

/** reads one byte from given queue
 *
 * Note: This function is not protected against concurrent calls from different tasks!
 * Note: It is safe to have one single task call push and another single task call pull on the same queue.
 * Note: It is not safe to have >1 task push/ pull on same queue!
 *
 * @param QueueHandle  handle of already created queue
 * @param Pointer      pointer address will be written to this address (NULL if queue is empty)
 * @return             == tqe_OK: Entry has been successfully pulled
 */
ttc_queue_error_e ttc_queue_pointer_pull_front(ttc_QueueHandle_Pointers_t QueueHandle, void** Pointer);

//}Function prototypes

#endif
