/*{ ttc_string.c ***********************************************
 *
 * Written by Gregor Rebel 2010-2012
 *
 * Architecture independent support for operations on zero terminated character sequences (strings).
 * 
}*/
#include "ttc_string.h"

//{ global variables *******************************************

// central function providing STDOUT like output of memory blocks
// Note: do not call this function directly, use ttc_string_block() instead!
// Note: do not write to this variable directly, use ttc_string_register_stdout() instead!
//
void (*_ttc_string_stdout_block)(ttc_memory_block_t* Block) = NULL;

// central function providing STDOUT like output of strings
// Note: do not call this function directly, use
//       ttc_string_print()/ ttc_string_printf() instead!
// Note: do not write to this variable directly, use ttc_string_register_stdout() instead!
//
void (*_ttc_string_stdout)(const u8_t* String, Base_t MaxSize) = NULL;

//}global variables
//{ functions *************************************************

void ttc_string_strip(u8_t** Start, Base_t* Size) {
    Assert_String(Start, ec_InvalidArgument);
    Assert_String(Size,  ec_InvalidArgument);

    u8_t* Reader   = *Start;
    Base_t NewSize = *Size;

    // remove leading whitespaces
    while ( (NewSize > 0) && ttc_string_is_whitechar(*Reader) ) {
        Reader++;
        NewSize--;
    }
    *Start = Reader;

    // remove trailing whitespaces
    Reader = *Start + NewSize - 1;
    while ( (NewSize > 0) && ttc_string_is_whitechar(*Reader) ) {
        Reader--;
        NewSize--;
    }
    *Size = NewSize;
}
BOOL ttc_string_is_whitechar(u8_t ASCII) {
    if (ASCII < 32) return TRUE;

    return FALSE;
}
Base_t ttc_string_copy(u8_t* Target, u8_t* Source, Base_t MaxSize) {
    Assert_String(Target, ec_InvalidArgument);
    Assert_String(Source, ec_InvalidArgument);

    Base_t BytesCopied = 0;
    u8_t C = 0;

    MaxSize--; // reserve 1 byte for string terminator

    while ( (BytesCopied < MaxSize) && ( (C = *Source++) != 0 ) ) {
        *Target++ = C;
        BytesCopied++;
    }
    *Target++ = 0;          // terminate string in Target[]

    return BytesCopied;
}
void ttc_string_print(const u8_t *String, Base_t MaxSize) {
    Assert_String(String, ec_InvalidArgument);

    if (_ttc_string_stdout) {
        _ttc_string_stdout(String, MaxSize);
    }
}
void ttc_string_printf(const char *Format, ...) {
    Assert_String(Format, ec_InvalidArgument);

    static u8_t* String    = NULL;
    static u8_t* WriterEnd = NULL;

#ifdef RT_SCHEDULER
    // Protect single String-buffer from concurrent calls
    static ttc_MutexHandle Lock = NULL;
#endif

    if (String == NULL) { // first call: allocate buffer
        Assert_String(String = ttc_memory_alloc(TTC_STRING_PRINTF_BUFFERSIZE), ec_Malloc);
        WriterEnd = String + TTC_STRING_PRINTF_BUFFERSIZE - 2; // reserve 1 byte for string terminator
#ifdef RT_SCHEDULER
        Lock = ttc_mutex_create();
#endif
    }
#ifdef RT_SCHEDULER
    while (ttc_mutex_lock(Lock, -1) != tqe_OK);
#endif

    va_list Arguments;
    // this is a protected function: use of static variables reduces stack usage
    static u8_t *Writer;        Writer        = String;
    static u8_t *Reader_Format; Reader_Format = (u8_t*) Format;

    va_start(Arguments, Format);
    static u8_t C; C = 0;
    Base_signed_t Value;

    for(; ( (C = *Reader_Format) != 0 ) && (Writer < WriterEnd); Reader_Format++)
    {
        if(C != '%') {
            *Writer++ = C;
            continue;
        }
        static u8_t FormatChar; FormatChar = *++Reader_Format;
        switch(FormatChar)
        {
        case 'c':
            *Writer++ = (u8_t) va_arg(Arguments, int);
            break;

        case 'd':
        case 'i':
            Value = va_arg(Arguments, Base_signed_t);
            Writer += ttc_itoa(Value, Writer, WriterEnd - Writer);
            break;

        case 's':
            Writer += ttc_string_copy(Writer, (u8_t*) va_arg(Arguments, char *), WriterEnd - Writer);
            break;

        case 'x':
            Value = va_arg(Arguments, Base_t);
            Writer += ttc_xtoa(Value, Writer, WriterEnd - Writer);
            break;

        default:
            *Writer++ = FormatChar;
            break;
        }
    }
    *Writer = 0;

    va_end(Arguments);
    ttc_string_print(String, (Base_t) (Writer - String));

#ifdef RT_SCHEDULER
    ttc_mutex_unlock(Lock);
#endif
}
u8_t* ttc_string_snprintf(u8_t* Buffer, Base_t Size, const char* Format, ...) {
    Assert_String(Format, ec_InvalidArgument);

    va_list Arguments;
    u8_t* Writer        = (u8_t*) Buffer;
    u8_t* WriterEnd     = (u8_t*) Buffer + Size;
    u8_t *Reader_Format = (u8_t*) Format;

    va_start(Arguments, Format);
    u8_t C = 0;
    Base_signed_t Value;

    for(; ( (C = *Reader_Format) != 0 ) && (Writer < WriterEnd); Reader_Format++)
    {
        if(C != '%') {
            *Writer++ = C;
            continue;
        }
        u8_t FormatChar = *++Reader_Format;
        switch(FormatChar)
        {
        case 'c':
            *Writer++ = (u8_t) va_arg(Arguments, int);
            break;

        case 'd':
        case 'i':
            Value = va_arg(Arguments, Base_signed_t);
            Writer += ttc_itoa(Value, Writer, WriterEnd - Writer);
            break;

        case 's':
            Writer += ttc_string_copy(Writer, (u8_t*) va_arg(Arguments, char *), WriterEnd - Writer);
            break;

        case 'x':
            Value = va_arg(Arguments, Base_t);
            Writer += ttc_xtoa(Value, Writer, WriterEnd - Writer);
            break;

        default:
            *Writer++ = FormatChar;
            break;
        }
    }
    if (Writer < WriterEnd)
        *Writer = 0;
    else
        *(Writer-1) = 0;

    va_end(Arguments);

    return Writer;
}
void ttc_string_print_block(ttc_memory_block_t* Block) {
    Assert_String(Block, ec_InvalidArgument);

    if (_ttc_string_stdout_block) {
        _ttc_string_stdout_block(Block);
    }
}
void ttc_string_register_stdout(void (*StdoutBlock)(ttc_memory_block_t*), void (*StdoutString)(const u8_t*, Base_t) ) {
    _ttc_string_stdout_block = StdoutBlock;
    _ttc_string_stdout       = StdoutString;
}
u8_t ttc_itoa(Base_signed_t Value, u8_t* Buffer, Base_t MaxSize) {
    u8_t BytesWritten = 0;

    if (Value == 0) { // Value == 0
        if (BytesWritten >= MaxSize)
            return BytesWritten;  // Buffer is full

        *Buffer++ = '0';
        return 1;
    }
    if (Value < 0)  { // print minus
        if (BytesWritten >= MaxSize)
            return BytesWritten;  // Buffer is full

        *Buffer++ = '-';
        Value *= -1;
        BytesWritten++;
    }

#if TARGET_DATA_WIDTH==32
    Base_t Base = 1000000000;
#endif
#if TARGET_DATA_WIDTH==16
    Base_t Base = 10000;
#endif
#if TARGET_DATA_WIDTH==8
    Base_t Base = 100;
#endif
    u8_t Digit;

    while ( (Base > 0) && (Base > Value) ) // skip leading zeroes
        Base /= 10;

    while (Base > 0) {
        if (BytesWritten >= MaxSize)
            return BytesWritten;  // Buffer is full

        if (Value >= Base) {
            Digit = Value / Base;
            Value -= Digit * Base;
        }
        else
            Digit = 0;

        *Buffer++ = '0' + Digit;
        BytesWritten++;

        Base /= 10;
    }
    return BytesWritten;
}
u8_t ttc_xtoa(Base_t Value, u8_t* Buffer, Base_t MaxSize) {
    u8_t BytesWritten = 0;

    if (Value == 0) { // Value == 0
        if (BytesWritten >= MaxSize)
            return BytesWritten;  // Buffer is full

        *Buffer++ = '0';
        return 1;
    }

#if TARGET_DATA_WIDTH==32
    Base_t Base = 0xf0000000;
    u8_t  Shift = 7 * 4;
#endif
#if TARGET_DATA_WIDTH==16
    Base_t Base = 0xf000;
    u8_t  Shift = 3 * 4;
#endif
#if TARGET_DATA_WIDTH==8
    Base_t Base = 0xf0;
    u8_t  Shift = 1 * 4;
#endif
    u8_t Digit;

    while ( (Base > 0) && ( (Base & Value) == 0) ) { // skip leading zeroes
        Base >>= 4;
        Shift -= 4;
    }

    while (Base > 0) {
        if (BytesWritten >= MaxSize)
            return BytesWritten;  // Buffer is full

        if (Value & Base)
            Digit = (Value & Base) >> Shift;
        else
            Digit = 0;

        if (Digit < 10)
            *Buffer++ = '0' + Digit;
        else
            *Buffer++ = 'a' + Digit - 10;

        BytesWritten++;
        Base >>= 4;
        Shift -= 4;
    }
    return BytesWritten;
}
u8_t ttc_byte2uint(u8_t Value, u8_t* Buffer) {

    if (Value >= 100) {
        u8_t Digit = Value / 100;
        *Buffer++ = '0' + Digit;
        Value -= Digit * 100;

        Digit = Value / 10;
        *Buffer++ = '0' + Digit;
        Value -= Digit * 10;

        *Buffer++ = '0' + Value;
        return 3;
    }
    if (Value >= 10) {
        u8_t Digit = Value / 10;
        *Buffer++ = '0' + Digit;
        Value -= Digit * 10;

        *Buffer++ = '0' + Value;
        return 2;
    }

    *Buffer++ = '0' + Value;
    return 1;
}
u8_t ttc_byte2int(s8_t Value, u8_t* Buffer) {
    if (Value < 0) { // negative value
        *Buffer++ = '-';
        return 1 + ttc_byte2uint((u8_t) 0x7f & Value, Buffer);
    }
    return ttc_byte2uint((u8_t) Value, Buffer);
}
void ttc_byte2hex(u8_t* Buffer, u8_t Byte) {
    u8_t Nibble = Byte & 0xf;

    if (Nibble < 10)
        *Buffer++ = '0' + Nibble;
    else
        *Buffer++ = 'a' - 10 + Nibble;

    Nibble = Byte >> 4;
    if (Nibble < 10)
        *Buffer = '0' + Nibble;
    else
        *Buffer = 'a' - 10 + Nibble;
}
u8_t strnlen__8(const char* String, u8_t MaxLen) {
    u8_t Length = 0;

    while ( (MaxLen > 0) && (*String != 0) ) {
        MaxLen--;
        String++;
        Length++;
    }
    return Length;
}
u16_t strnlen__16(const char* String, u16_t MaxLen) {
    u16_t Length = 0;

    while ( (MaxLen > 0) && (*String != 0) ) {
        MaxLen--;
        String++;
        Length++;
    }
    return Length;
}
u32_t strnlen__32(const char* String, u32_t MaxLen) {
    u32_t Length = 0;

    while ( (MaxLen > 0) && (*String != 0) ) {
        MaxLen--;
        String++;
        Length++;
    }
    return Length;
}

//}functions
