/*{ ExampleThreading::.c ************************************************

 Example code demonstrating how to use inter thread communication.
 
}*/

#include "example_threading_queues.h"


#ifdef EXTENSION_500_ttc_gfx
  const char* Example   = "Example";
  const char* Threading = "Threading";
  const char* Queues    = "Queues";

  // configuration data of connected display
  ttc_gfx_generic_t* DisplayConfig = NULL;
#endif

//{ Function definitions *************************************************

void threading_init() {
#ifdef EXTENSION_500_ttc_gfx
    u8_t MaxDisplay = ttc_gfx_get_max_display_index();
    Assert(MaxDisplay > 0, ec_DeviceNotFound); // no displays defined by board makefile!
    DisplayConfig = ttc_gfx_get_configuration(1);
    Assert(DisplayConfig, ec_DeviceNotFound);

    // set color to use for first screen clear
    ttc_gfx_set_color_bg24(GFX_COLOR24_GREEN);
    ttc_gfx_set_color_fg24(GFX_COLOR24_WHITE);
    ttc_gfx_init(1);

    // initialize gfx for use with printf()
    ttc_gfx_stdout_set(1);
    ttc_string_register_stdout(ttc_gfx_stdout_print_block, ttc_gfx_stdout_print_string);

    ttc_gfx_text_cursor_set( ttc_gfx_text_center(Example, -1), 3);
    printf(Example);
    ttc_gfx_text_cursor_set( ttc_gfx_text_center(Threading, -1), 4);
    printf(Threading);
#endif
}
void threading_start() {
    threading_init();

    static et_ThreadInfo_t ThreadInfo; // must be static to survive end of this function!

    // create queue Producer -> Consumer
    ThreadInfo.QueueProducer2Consumer = ttc_queue_create(10, 1);

    // create queue Producer/Consumer -> Print
    ThreadInfo.QueuePrint = ttc_queue_create(5, sizeof(et_PrintJob_t));

    Assert(NULL != ThreadInfo.QueueProducer2Consumer, ec_Malloc);
    Assert(NULL != ThreadInfo.QueuePrint, ec_Malloc);

    BOOL Task1 = ttc_task_create(taskProducer,                    // function to start as thread
                                 "taskProducer",                  // thread name (just for debugging)
                                 4 * TTC_TASK_MINIMAL_STACK_SIZE, // stack size
                                 (void *) &ThreadInfo,            // passed as argument to taskValue()
                                 1,                               // task priority (higher values mean more process time)
                                 NULL                             // can return a handle to created task
                                 );

    BOOL Task2 = ttc_task_create(taskConsumer,                    // function to start as thread
                                 "taskConsumer",                  // thread name (just for debugging)
                                 4 * TTC_TASK_MINIMAL_STACK_SIZE, // stack size
                                 (void *) &ThreadInfo,            // passed as argument to taskValue()
                                 1,                               // task priority (higher values mean more process time)
                                 NULL                             // can return a handle to created task
                                 );

    BOOL Task3 = ttc_task_create(taskPrint,                       // function to start as thread
                                 "taskPrint",                     // thread name (just for debugging)
                                 4 * TTC_TASK_MINIMAL_STACK_SIZE, // stack size
                                 (void *) &ThreadInfo,            // passed as argument to taskValue()
                                 1,                               // task priority (higher values mean more process time)
                                 NULL                             // can return a handle to created task
                                 );

    Assert(Task1, ec_Malloc);
    Assert(Task2, ec_Malloc);
    Assert(Task3, ec_Malloc);

    for (int Row = 0; Row < DisplayConfig->TextRows; Row++)
        queuePrintJob(ThreadInfo.QueuePrint, 0, Row, "%x", 0xf & Row);
    for (int Col = 0; Col < DisplayConfig->TextColumns; Col++)
        queuePrintJob(ThreadInfo.QueuePrint, Col, 0, "%x", 0xf & Col);

    queuePrintJob(ThreadInfo.QueuePrint, ttc_gfx_text_center(Queues, -1), 5, Queues, 0);
    Base_t BytesFree = ttc_memory_get_free_size();
    queuePrintJob(ThreadInfo.QueuePrint, 2, DisplayConfig->TextRows - 1, "free: %iB", BytesFree);
}
void taskProducer(void *pvParameters) {
    et_ThreadInfo_t* ThreadInfo = (et_ThreadInfo_t*) pvParameters;
    Assert(NULL != ThreadInfo, ec_NULL);
    Assert(NULL != ThreadInfo->QueueProducer2Consumer, ec_InvalidArgument);

    queuePrintJob(ThreadInfo->QueuePrint, 2, 7, "Producer>", 0);

#ifdef TTC_LED1
    ttc_gpio_init(TTC_LED1, tgm_output_push_pull);
#endif

    u8_t Value = 0; //(u16_t) rand() & 0xfff;
    while (1) {
        while (ttc_queue_push_back(ThreadInfo->QueueProducer2Consumer, &Value, -1) != tqe_OK);

        queuePrintJob(ThreadInfo->QueuePrint, 11, 7, "%i ", Value);

        Base_t QueueSize = ttc_queue_amount_waiting(ThreadInfo->QueueProducer2Consumer);
        queuePrintJob(ThreadInfo->QueuePrint, 2, 9, "|Q| = %i ", (u16_t) QueueSize);
#ifdef TTC_LED1
        for (int I = Value; I >= 0; I--) {
            ttc_gpio_set(TTC_LED1);
            mSleep(100);
            ttc_gpio_clr(TTC_LED1);
            mSleep(100);
        }
#endif
        mSleep(500);

        Value++;
        if (Value == 10) Value = 0;
    }
}
void taskConsumer(void *pvParameters) {
    et_ThreadInfo_t* ThreadInfo = (et_ThreadInfo_t*) pvParameters;
    Assert(NULL != ThreadInfo, ec_NULL);
    Assert(NULL != ThreadInfo->QueueProducer2Consumer, ec_InvalidArgument);

    queuePrintJob(ThreadInfo->QueuePrint, 2, 8, "Consumer>", 0);
#ifdef TTC_LED2
    ttc_gpio_init(TTC_LED2, tgm_output_push_pull);
#endif

    u8_t ReceivedValue = 0;
    while (1) {
        if ( ttc_queue_pop_front(ThreadInfo->QueueProducer2Consumer, &ReceivedValue,  (portTickType) -1) == tqe_OK ) {
            Assert( ReceivedValue <= 100, ec_InvalidArgument );
            queuePrintJob(ThreadInfo->QueuePrint, 11, 8, "%i ", ReceivedValue);
#ifdef TTC_LED2
            for (int I = ReceivedValue; I >= 0; I--) {
                ttc_gpio_set(TTC_LED2);
                mSleep(100);
                ttc_gpio_clr(TTC_LED2);
                mSleep(100);
            }
#endif
            ttc_testpoint(); // test run successfull
            mSleep(600); // consumer is a little bit slower than producer
        }
    }
}
void queuePrintJob(ttc_QueueHandle PrintQueue, s8_t Column, s8_t Row, const char* Text, s32_t Value) {
    et_PrintJob_t PrintJob;
    PrintJob.Column = Column;
    PrintJob.Row    = Row;
    PrintJob.Value  = Value;
    strncpy(PrintJob.Text, Text, 10);

    while (ttc_queue_push_back(PrintQueue, &PrintJob, -1) != tqe_OK);
}
void taskPrint(void *pvParameters) {
    et_ThreadInfo_t* ThreadInfo = (et_ThreadInfo_t*) pvParameters;
    Assert(NULL != ThreadInfo, ec_InvalidArgument);

    while (1) {
        et_PrintJob_t PrintJob;
        if (ttc_queue_pop_front(ThreadInfo->QueuePrint, &PrintJob,  (portTickType) -1) == tqe_OK) {
#ifdef EXTENSION_500_ttc_gfx
            if ( (PrintJob.Column != -1) && (PrintJob.Row != -1) )
                ttc_gfx_text_cursor_set(PrintJob.Column, PrintJob.Row);
            printf(PrintJob.Text, PrintJob.Value);
#endif
        }
    }
}

//} Function definitions
